from typing import Tuple, Union

import jax.numpy as jnp

from photon_weave.state.expansion_levels import ExpansionLevel


def state_expand(
    state: Union[jnp.ndarray, int],
    current_expansion_level: ExpansionLevel,
    dimensions: int,
) -> Tuple[jnp.ndarray, ExpansionLevel]:
    """
    Returns the expanded state representation.
    If currently the state is represented as a label,
    it returns the state vector, if it is repersented
    as a state vector, then it is expanded into density
    matrix

    Parameters
    ----------
    state: Union[int, jnp.ndarray]
        State in the form of Label, State Vector or Density Matrix
    current_expansion_level: ExpansionLevel
        Current Expansion Level
    dimensions: int
        Dimensionality of the system

    Returns
    -------
    Tuple[jnp.ndarray, ExpansionLevel]
        Returns expanded state representation with the new representation
        level

    Notes
    -----
    In the case of Polarization, it can expand from State Vector onward
    """
    assert isinstance(dimensions, int)
    assert isinstance(current_expansion_level, ExpansionLevel)
    if dimensions < 0:
        raise ValueError("Dimensions must be larger than 0")
    new_state: Union[jnp.ndarray]
    match current_expansion_level:
        case ExpansionLevel.Label:
            if not isinstance(state, int):
                raise ValueError("Could not expand state, where label is not int type")
            assert state >= 0
            new_state = jnp.zeros(dimensions, dtype=jnp.complex128)
            new_state = new_state.at[state].set(1)
            new_state = new_state[:, jnp.newaxis]
            new_expansion_level = ExpansionLevel.Vector

        case ExpansionLevel.Vector:
            assert isinstance(state, jnp.ndarray)
            assert state.shape == (dimensions, 1)

            new_state = jnp.outer(state.flatten(), jnp.conj(state.flatten()))
            new_expansion_level = ExpansionLevel.Matrix
        case ExpansionLevel.Matrix:
            assert isinstance(state, jnp.ndarray)
            new_state = state
            new_expansion_level = current_expansion_level
        case _:
            raise ValueError("Something went wrong")
    return new_state, new_expansion_level


def state_contract(
    state: Union[int, jnp.ndarray],
    current_expansion_level: ExpansionLevel,
    tol: float = 1e-6,
) -> Tuple[Union[int, jnp.ndarray], ExpansionLevel, bool]:
    """
    Returns contracted state representation if possible

    Parameters
    ----------
    state: Union[int, jnp.ndarray]
        Current state, which should be contracted
    current_expansion_level: ExpansionLevel
        Current expansion level of the state
    tol: float
        Tolerance, used when contracting from matrix to vector representation

    Returns
    -------
    Tuple[Union[jnp.ndarray,int], ExpansionLevel, bool]
        Returns a tuple of the new state, new representation expantion level
        and success flag. Is process was succesfull, then the success is True

    Notes
    -----
    Needs to be handled carefully in the case of Polarization, because
    Polarization uses Enum when representing Label states
    """

    assert isinstance(current_expansion_level, ExpansionLevel)

    success = False
    new_state = state
    new_expansion_level = current_expansion_level

    match current_expansion_level:
        case ExpansionLevel.Matrix:
            assert isinstance(state, jnp.ndarray)
            assert state.shape[0] == state.shape[1]
            state_squared = jnp.matmul(state, state)
            state_trace = jnp.trace(state_squared)
            if jnp.abs(state_trace - 1) < tol:
                eigenvalues, eigenvectors = jnp.linalg.eigh(state)
                pure_state_index = jnp.argmax(eigenvalues)
                new_state = eigenvectors[:, pure_state_index].reshape(-1, 1)
                # Removing the global phase
                assert isinstance(new_state, jnp.ndarray)
                phase = jnp.exp(-1j * jnp.angle(new_state[0]))
                new_state = new_state * phase
                new_expansion_level = ExpansionLevel.Vector
                success = True

        case ExpansionLevel.Vector:
            assert isinstance(state, jnp.ndarray)
            assert state.shape[1] == 1
            ones = jnp.where(state == 1)[0]
            if ones.size == 1:
                new_state = int(ones[0])
                new_expansion_level = ExpansionLevel.Label
                success = True

    return (new_state, new_expansion_level, success)



import importlib
from enum import Enum
from typing import Any, List, Union

import jax.numpy as jnp
from jax.scipy.linalg import expm

from photon_weave._math.ops import (
    annihilation_operator,
    controlled_not_operator,
    controlled_swap_operator,
    controlled_z_operator,
    creation_operator,
    swap_operator,
)
from photon_weave.extra import interpreter
from photon_weave.state.base_state import BaseState
from photon_weave.state.expansion_levels import ExpansionLevel


class CompositeOperationType(Enum):
    r"""
    CompositeOperationType

    Constructs an operator, which acts on multiple spaces

    NonPolarizingBeamSplitter
    -------------------------
    Constructs a non-polarizing  beam splitter operator that acts on two Fock spaces.
    The operator is represented by a unitary transformation that mices the two modes.
    The constructed operator is of the form:
    .. math::

        \hat U_{BS} = e^{i\theta(\hat a^\dagger \hat b + \hat a \hat b^\dagger)}

    For a 50/50 beam splitter, :math:`\theta = \frac{\pi}{4}\`, which leas to equal
    mixing of the two modes.

    CNOT (CXPolarization)
    -----------------------
    Constructs a CNOT operator operating on two polarization states. First state
    provided is control and the second is target

    .. math::

        \hat{CX} = \begin{bmatrix}1&0&0&0\\0&0&0&1\\0&0&1&0\\0&1&0&0\end{bmatrix}

    >>> op = Operation(CompositeOperationType.CXPolarization)
    >>> ce.apply_operation(op, control_pol, target_pol)

    CZ (CZPolarization)
    -----------------------
    Constructs a controlled-Z opreator operating on two polarization states.
    First state provided is control and the second is target.
    .. math::

        \hat{CZ} = \begin{bmatrix}1&0&0&0\\0&1&0&0\\0&0&1&0\\0&0&0&-1\end{bmatrix}

    >>> op = Operation(CompositeOperationType.CZPolarization)
    >>> ce.apply_operation(op, control_pol, target_pol)

    SWAP (SwapPolarization)
    -----------------------
    Constructs a SWAP operation, swaping the states of two provided polarization
    states.

    .. math::
        \hat{SWAP} = \begin{bmatrix}1&0&0&0\\0&0&1&0\\0&1&0&0\\0&0&0&1\end{bmatrix}

    Example Usage:
    >>> op = Operation(CompositeOperationType.SwapPolarization)
    >>> ce.apply_operation(op, env1.polarization, env2.polarization)

    Controlled-SWAP (CSwapPolarization)
    -----------------------------------
    Constructs a Controlled-SWAP operation, conditionally swapping the states of
    two provided polarization states. First state is the control state and the
    next two states are target states.
    .. math::

        \hat{CSWAP} =
        \begin{bmatrix}
        1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
        0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
        0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\
        0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\
        0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 \\
        0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
        0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
        0 & 0 & 0 & 0 & 0 & 0 & 0 & 1
        \end{bmatrix}

    Example Usage:
    >>> op = Operation(CompositeOperationType.SwapPolarization)
    >>> ce.apply_operation(op, env1.polarization, env2.polarization)

    Expression
    ----------
    Constucts an operator based on the expression provided. Alongside expression
    also list of state types needs to be provided together with the context.
    Context needs to be a dictionary of operators, where the keys are strings used in
    the expression and values are lambda functions, expecting one argument: dimensions.

    An example of context and operator usage
    >>> context = {
    >>>    "a_dag": lambda dims: creation_operator(dims[0])
    >>>    "a":     lambda dims: annihilation_operator(dims[0])
    >>>    "n":     lambda dims: number_operator(dims[0])
    >>> }
    >>> op = Operation(CompositeOperationType.Expression,
    >>>                expr=("expm"("s_mult", 1j,jnp.pi, "a"))),
    >>>                state_types=(Fock,), # Is applied to only one fock space
    >>>                context=context)
    >>> ce.apply_operation(op, fock_in_ce)

    It is IMPORTANT to correctly use the indexing in the context dictionary, when using
    dim[0], this operator will be used for the first state. And the dimensions of first
    state will be passed to the lambda function.

    It is also IMPORTANT to correctly write the expression as expression doesn't have a
    dimension asserting functionality, meaning that if you operate on two states, you
    need to correctly kron the operators to achieve correct operator dimensionality.

    For example if you which to operate on two Fock spaces like in beam splitter example
    you need to produce the following context and expression:
    >>> context = {
    >>>     'a_dag': lambda dims: creation_opreator(dims[0])
    >>>     'a':     lambda dims: annihilation_opreator(dims[0])
    >>>     'b_dag': lambda dims: creation_opreator(dims[1])
    >>>     'b':     lambda dims: annihilation_opreator(dims[1])
    >>> }
    >>> expression = ('expm',
    >>>                ('s_mult',
    >>>                  1j, jnp.pi/4,
    >>>                  ('add',
    >>>                    ('kron', a_dag, b),
    >>>                    ('kron', a, b_dag))))
    >>> op = Operation(
    >>>         CompositeOperationType.Expresion,
    >>>         expr=expression,
    >>>         state_types=(Fock, Fock),
    >>>         context=context)
    >>> ce.apply_operation(op, fock1, fock2)

    In this example 'a_dag' and 'a' will correspond to the fock1 state and
    'b_dag' and 'b' will correspond to the fock2 state. Notice also that the
    operators are always kronned in the order that corresponds to the usage later
    on. Application of the operation already reorders the states in a product state,
    such that their order corresponds to the order called in the apply_operation
    method call.
    """

    NonPolarizingBeamSplitter = (
        True,
        ["eta"],
        ["Fock", "Fock"],
        ExpansionLevel.Vector,
        1,
    )  # type: ignore
    CXPolarization = (
        True,
        [],
        ["Polarization", "Polarization"],
        ExpansionLevel.Vector,
        2,
    )  # type: ignore
    SwapPolarization = (
        True,
        [],
        ["Polarization", "Polarization"],
        ExpansionLevel.Vector,
        3,
    )  # type: ignore
    CSwapPolarization = (
        True,
        [],
        ["Polarization" for _ in range(3)],
        ExpansionLevel.Vector,
        4,
    )  # type: ignore
    CZPolarization = (
        True,
        [],
        ["Polarization" for _ in range(2)],
        ExpansionLevel.Vector,
        5,
    )  # type: ignore
    Expression = (
        True,
        ["expr", "state_types", "context"],
        [],
        ExpansionLevel.Vector,
        6,
    )  # type: ignore

    def __init__(
        self,
        renormalize: bool,
        required_params: List[str],
        expected_base_state_types: List[BaseState],
        required_expansion_level: ExpansionLevel,
        op_id: int,
    ) -> None:
        self.renormalize = renormalize
        self.required_params = required_params
        self.expected_base_state_types = expected_base_state_types
        self.required_expansion_level = required_expansion_level

    def update(self, **kwargs: Any) -> None:
        """
        Updates the required_base_types in the case of Expression type

        Parameters
        ----------
        **kwargs: Any
            kwargs, where "state_types" is included
        """
        Fock = importlib.import_module("photon_weave.state.fock").Fock
        Polarization = importlib.import_module(
            "photon_weave.state.polarization"
        ).Polarization
        CustomState = importlib.import_module(
            "photon_weave.state.custom_state"
        ).CustomState
        if self is CompositeOperationType.Expression:
            self.expected_base_state_types = list(kwargs["state_types"])
        for i, state_type in enumerate(self.expected_base_state_types):
            if state_type == "Fock":
                self.expected_base_state_types[i] = Fock
            elif state_type == "Polarization":
                self.expected_base_state_types[i] = Polarization
            elif state_type == "CustomState":
                self.expected_base_state_types[i] = CustomState

    def compute_operator(self, dimensions: List[int], **kwargs: Any) -> jnp.ndarray:
        """
        Generates the operator for this operation, given
        the dimensions

        Parameters
        ----------
        dimensions: List[int]
            The dimensions of the spaces
        **kwargs: Any
            List of key word arguments for specific operator types
        """
        match self:
            case CompositeOperationType.NonPolarizingBeamSplitter:
                a = creation_operator(dimensions[0])
                a_dagger = annihilation_operator(dimensions[0])
                b = creation_operator(dimensions[1])
                b_dagger = annihilation_operator(dimensions[1])

                operator = jnp.kron(a_dagger, b) + jnp.kron(a, b_dagger)
                return expm(1j * kwargs["eta"] * operator)
            case CompositeOperationType.CXPolarization:
                return controlled_not_operator()
            case CompositeOperationType.SwapPolarization:
                return swap_operator()
            case CompositeOperationType.CSwapPolarization:
                return controlled_swap_operator()
            case CompositeOperationType.CZPolarization:
                return controlled_z_operator()
            case CompositeOperationType.Expression:
                return interpreter(kwargs["expr"], kwargs["context"], dimensions)
        raise ValueError("Operation Type not recognized")

    def compute_dimensions(
        self,
        num_quanta: Union[int, List[int]],
        states: Union[jnp.ndarray, List[jnp.ndarray]],
        threshold: float = 1 - 1e6,
        **kwargs: Any,
    ) -> List[int]:
        """
        Compute the dimensions for the operator. Application of the
        operator could change the dimensionality of the space. For
        example creation operator, would increase the dimensionality
        of the space, if the prior dimensionality doesn't account
        for the new particle.

        Parameters
        ----------
        num_quanta: int
            Number of particles in the space currently. In other words
            highest basis with non_zero probability in the space
        state: jnp.ndarray
            Traced out state, usede for final dimension estimation
        threshold: float
            Minimal amount of state that has to be included in the
            post operation state
        **kwargs: Any
            Additional parameters, used to define the operator

        Returns
        -------
        int
           New number of dimensions

        Notes
        -----
        This functionality is called before the operator is computed, so that
        the dimensionality of the space can be changed before the application
        and the dimensionality of the operator and space match
        """
        from photon_weave.state.fock import Fock

        match self:
            case CompositeOperationType.NonPolarizingBeamSplitter:
                dim = int(jnp.sum(jnp.array(num_quanta))) + 1
                return [dim, dim]
            case CompositeOperationType.CXPolarization:
                return [2, 2]
            case CompositeOperationType.SwapPolarization:
                return [2, 2]
            case CompositeOperationType.CSwapPolarization:
                return [2, 2, 2, 2]
            case CompositeOperationType.CZPolarization:
                return [2, 2]
            case CompositeOperationType.Expression:
                assert isinstance(num_quanta, list)
                dims = []
                for i, s in enumerate(self.expected_base_state_types):
                    if s is Fock:
                        dims.append(num_quanta[i] + 2)
                    else:
                        dims.append(num_quanta[i])
                return dims
        raise ValueError("Operation Type not recognized")



"""
This module implements Custom State

Custom state can be used to implement various quantum system
such as qubits or quantum dots
"""

import uuid
from typing import Dict, List, Optional, Tuple, Union

import jax
import jax.numpy as jnp

from photon_weave._math.ops import apply_kraus, kraus_identity_check
from photon_weave.operation import CustomStateOperationType, Operation
from photon_weave.photon_weave import Config
from photon_weave.state.base_state import BaseState
from photon_weave.state.composite_envelope import CompositeEnvelope
from photon_weave.state.expansion_levels import ExpansionLevel

from .utils.measurements import measure_matrix, measure_vector
from .utils.operations import apply_operation_matrix, apply_operation_vector
from .utils.routing import route_operation
from .utils.state_transform import state_contract, state_expand


class CustomState(BaseState):
    def __init__(self, dimensions: int):
        self.uid = uuid.uuid4()
        # Custom state can only be separate or part of composite envelope
        self.index: Optional[Tuple[int, int]] = None
        self._dimensions_set = False
        self.dimensions = dimensions
        # Initialize in the |0> state
        self.state: Optional[Union[int, jnp.ndarray]] = 0
        self.expansion_level = ExpansionLevel.Label
        self.composite_envelope: Optional["CompositeEnvelope"] = None

    @property
    def uid(self) -> Union[str, uuid.UUID]:
        return self._uid

    @uid.setter
    def uid(self, uid: Union[uuid.UUID, str]) -> None:
        self._uid = uid

    @property
    def dimensions(self) -> int:
        return self._dimensions

    @dimensions.setter
    def dimensions(self, dimensions: int) -> None:
        if self._dimensions_set:
            raise ValueError("Dimensions can only be set once")
        self._dimensions = dimensions
        self._dimensions_set = True

    @property
    def index(self) -> Union[None, int, Tuple[int, int]]:
        assert not isinstance(self._index, int)
        return self._index

    @index.setter
    def index(self, index: Union[None, int, Tuple[int, int]]) -> None:
        assert not isinstance(index, int)
        self._index = index

    @route_operation()
    def expand(self) -> None:
        """
        Expands the state from label to vector and from vector to matrix

        Notes
        -----
        Method is decorated with route_operation. If the state is
        contained in the product state, the corresponding operation
        will be executed in the state container, which contains this
        stat.
        """
        assert self.state is not None
        assert isinstance(self.expansion_level, ExpansionLevel)
        self.state, self.expansion_level = state_expand(
            self.state, self.expansion_level, self.dimensions
        )

    @route_operation()
    def contract(
        self, final: ExpansionLevel = ExpansionLevel.Label, tol: float = 1e-6
    ) -> None:
        """
        Attempts to contract the representation to the level defined in final argument.

        Parameters
        ----------
        final: ExpansionLevel
            Expected expansion level after contraction
        tol: float
            Tolerance when comparing matrices

        """
        assert isinstance(self.expansion_level, ExpansionLevel)
        assert isinstance(self.state, jnp.ndarray)

        success = True
        while self.expansion_level > final and success:
            self.state, self.expansion_level, success = state_contract(
                self.state, self.expansion_level
            )

    @property
    def _measured(self) -> bool:
        return False

    @_measured.setter
    def _measured(self, measured: bool) -> None:
        """
        Configured property, without effect
        Custom state cannot be destroyed, so
        measured is without effect
        """
        pass

    def _set_measured(self) -> None:
        pass

    def extract(self, index: Union[int, Tuple[int, int]]) -> None:
        assert isinstance(index, tuple)
        assert len(index) == 2
        self.index = index
        self.state = None

    def set_index(self, minor: int = -1, major: int = -1) -> None:
        if minor == -1 and major == -1:
            self.index = None
        elif major >= 0 and minor >= 0:
            self.index = (major, minor)
        else:
            raise ValueError(
                "Either set both parameters (minor, major) or none of them"
            )

    @route_operation()
    def measure(
        self, separate_measurement: bool = False, destructive: bool = True
    ) -> Dict[BaseState, int]:
        """
        Measures the state and returns an outcome in a dict

        Parameters
        ----------
        separate_measurement: bool
            Doesn't have any effect, implemented to not defy the signature
            of the superclass
        destructive: bool
            Doesn't have any effect, implemented to not defy the signature
            of the superclass

        Returns
        -------
        Dict[BaseState, int]
            Dictionary of outcomes, in this case only one outcome

        Notes
        -----
        Method is decorated with route_operation. If the state is
        contained in the product state, the corresponding operation
        will be executed in the state container, which contains this
        stat.
        """
        match self.expansion_level:
            case ExpansionLevel.Label:
                assert isinstance(self.state, int)
                return {self: self.state}
            case ExpansionLevel.Vector:
                assert isinstance(self.state, jnp.ndarray)
                outcomes, post_measurement_state = measure_vector(
                    [self], [self], self.state
                )
            case ExpansionLevel.Matrix:
                assert isinstance(self.state, jnp.ndarray)
                outcomes, post_measurement_state = measure_matrix(
                    [self], [self], self.state
                )

        self.state = outcomes[self]
        self.expansion_level = ExpansionLevel.Label

        return outcomes

    def measure_POVM(
        self,
        operators: List[jnp.ndarray],
        destructive: bool = True,
        partial: bool = False,
    ) -> Tuple[int, Dict[BaseState, int]]:
        """
        Positive Operation-Valued Measurement

        Parameters
        ----------
        operators: List[jnp.ndarray]
            List of the POVM measurement operators
        destructive: bool
            Does not have an effect on custom state, implemented only to satisfy the
            superclass method signature
        partial: bool
            Does not have an effect on custom state, implemented only to satisfy the
            superclass method signature

        Returns
        -------
        Tuple[int, Dict[BaseState, int]]
            Tuple, where first value is the outcome of this POVM measurement,
            second is empty, used when envelope parts are measured
        """
        if self.index is not None:
            assert isinstance(self.composite_envelope, CompositeEnvelope)
            return self.composite_envelope.measure_POVM(operators, self)

        for op in operators:
            assert op.shape == (self.dimensions, self.dimensions)

        assert isinstance(self.expansion_level, ExpansionLevel)
        while self.expansion_level < ExpansionLevel.Matrix:
            self.expand()

        C = Config()

        assert isinstance(self.state, jnp.ndarray)
        probabilities = jnp.array(
            [jnp.trace(jnp.matmul(op, self.state)).real for op in operators]
        )
        probabilities = probabilities / jnp.sum(probabilities)

        key = C.random_key
        outcome = int(
            jax.random.choice(key, a=jnp.arange(len(operators)), p=probabilities)
        )
        self.state = jnp.matmul(
            operators[outcome], jnp.matmul(self.state, jnp.conj(operators[outcome].T))
        )
        self.state = self.state / jnp.trace(self.state)

        if C.contractions:
            self.contract()
        return (outcome, {})

    def apply_kraus(
        self,
        operators: List[jnp.ndarray],
        identity_check: bool = True,
    ) -> None:
        """
        Apply Kraus operators to the state. The State is automatically expanded to the
        density matrix representation.

        Parameters
        ----------
        operators: List[Union[np.ndarray, jnp.Array]]
            List of the operators
        identity_check: bool
            Signal to check whether or not the operators sum up to identity, True by
            default
        """
        if self.index is not None:
            assert isinstance(self.composite_envelope, CompositeEnvelope)
            self.composite_envelope.apply_kraus(operators, self)

        while self.expansion_level != ExpansionLevel.Matrix:
            self.expand()

        for op in operators:
            if op.shape != (self.dimensions, self.dimensions):
                raise ValueError(
                    f"Kraus operator has incorrect dimensions: {op.shape}, "
                    f"expected ({self.dimensions},{self.dimensions})"
                )

        if identity_check:
            if not kraus_identity_check(operators):
                raise ValueError("Invalid Kraus Channel")

        self.state = apply_kraus(self.state, operators)
        C = Config()
        if C.contractions:
            self.contract()

    @route_operation()
    def apply_operation(self, operation: Operation) -> None:
        """
        Applies an operation to the state. If state is in some product
        state, the operator is correctly routed to the specific
        state

        Parameters
        ----------
        operation: Operation
            Operation with operation type: FockOperationType

        Notes
        -----
        Method is decorated with route_operation. If the state is
        contained in the product state, the corresponding operation
        will be executed in the state container, which contains this
        stat.
        """
        assert isinstance(operation._operation_type, CustomStateOperationType)
        assert isinstance(self.expansion_level, ExpansionLevel)

        while self.expansion_level < operation.required_expansion_level:
            self.expand()

        # Consolidate the dimensions
        to = self.trace_out()
        assert isinstance(to, jnp.ndarray)
        operation.compute_dimensions(0, to)

        assert operation.operator.shape == (self.dimensions, self.dimensions)
        assert isinstance(self.state, jnp.ndarray)

        match self.expansion_level:
            case ExpansionLevel.Vector:
                self.state = apply_operation_vector(
                    [self], [self], self.state, operation.operator
                )
            case ExpansionLevel.Matrix:
                self.state = apply_operation_matrix(
                    [self], [self], self.state, operation.operator
                )

        if not jnp.any(jnp.abs(self.state) > 0):
            raise ValueError(
                "The state is entirely composed of zeros, "
                "is |0⟩ attempted to be annihilated?"
            )

        if operation.renormalize:
            self.state = self.state / jnp.linalg.norm(self.state)

        C = Config()
        if C.contractions:
            self.contract()



"""
Envelope
"""

# ruff: noqa: F401

from __future__ import annotations

import uuid
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union

import jax.numpy as jnp
import numpy as np
from jax import jit
from scipy.integrate import quad

from photon_weave._math.ops import (
    kraus_identity_check,
    num_quanta_matrix,
    num_quanta_vector,
)
from photon_weave.constants import C0, gaussian
from photon_weave.operation import (
    FockOperationType,
    Operation,
    PolarizationOperationType,
)
from photon_weave.photon_weave import Config
from photon_weave.state.expansion_levels import ExpansionLevel
from photon_weave.state.fock import Fock
from photon_weave.state.polarization import Polarization

from .utils.measurements import measure_matrix, measure_POVM_matrix, measure_vector
from .utils.operations import (
    apply_kraus_matrix,
    apply_operation_matrix,
    apply_operation_vector,
)
from .utils.representation import representation_matrix, representation_vector
from .utils.state_transform import state_contract, state_expand
from .utils.trace_out import trace_out_matrix, trace_out_vector

if TYPE_CHECKING:
    # from photon_weave.operation import Operation
    from photon_weave.state.composite_envelope import CompositeEnvelope
    from photon_weave.state.polarization import PolarizationLabel

    from .base_state import BaseState

jitted_kron = jit(jnp.kron)


class TemporalProfile(Enum):
    Gaussian = (gaussian, {"mu": 0, "sigma": 1, "omega": None})

    def __init__(self, func: Callable, params: Any):
        self.func = func
        self.parameters = params

    def with_params(self, **kwargs: Any) -> TemporalProfileInstance:
        params = self.parameters.copy()
        params.update(kwargs)
        return TemporalProfileInstance(self.func, params)


class TemporalProfileInstance:
    def __init__(self, func: Callable, params: Dict[Any, Any]) -> None:
        self.func = func
        self.params = params

    def get_function(self, t_a: float, omega_a: float) -> Callable:
        params = self.params.copy()
        params.update({"t_a": t_a, "omega": omega_a})

        return lambda t: self.func(t, **params)


default_temporal_profile = TemporalProfile.Gaussian.with_params(
    mu=0,
    sigma=42.45 * 10 ** (-15),  # 100 fs pulse
)


class Envelope:
    __slots__ = (
        "uid",
        "state",
        "_expansion_level",
        "_composite_envelope_id",
        "measured",
        "wavelength",
        "temporal_profile",
        "__dict__",
    )

    def __init__(
        self,
        wavelength: float = 1550,
        fock: Optional["Fock"] = None,
        polarization: Optional["Polarization"] = None,
        temporal_profile: TemporalProfileInstance = default_temporal_profile,
    ):
        from photon_weave.state.fock import Fock
        from photon_weave.state.polarization import Polarization

        self.uid: uuid.UUID = uuid.uuid4()
        self.fock = Fock() if fock is None else fock
        self.fock.envelope = self
        self.polarization = Polarization() if polarization is None else polarization
        self.polarization.envelope = self
        self._expansion_level: Optional[ExpansionLevel] = None
        self.composite_envelope_id: Optional[uuid.UUID] = None
        self.state: Optional[jnp.ndarray] = None
        self.measured = False
        self.wavelength = wavelength
        self.temporal_profile = temporal_profile

    @property
    def expansion_level(self) -> Optional[ExpansionLevel]:
        if self._expansion_level is not None:
            return self._expansion_level
        if self.fock.expansion_level == self.polarization.expansion_level:
            return self.fock.expansion_level
        return None

    @expansion_level.setter
    def expansion_level(self, expansion_level: ExpansionLevel) -> None:
        self._expansion_level = expansion_level
        self.fock.expansion_level = expansion_level
        self.polarization.expansion_level = expansion_level

    @property
    def dimensions(self) -> int:
        fock_dims = self.fock.dimensions
        pol_dims = self.polarization.dimensions
        return fock_dims * pol_dims

    def __repr__(self) -> str:
        if self.measured:
            return "Envelope already measured"
        if self.state is None:
            fock_repr = self.fock.__repr__().splitlines()
            pol_repr = self.polarization.__repr__().splitlines()
            # Maximum length accross fock repr
            max_length = max(len(line) for line in fock_repr)
            max_lines = max(len(fock_repr), len(pol_repr))

            fock_repr.extend([" " * max_length] * (max_lines - len(fock_repr)))
            pol_repr.extend([""] * (max_lines - len(pol_repr)))
            zipped_lines = zip(fock_repr, pol_repr)
            return "\n".join(f"{f_line} ⊗ {p_line}" for f_line, p_line in zipped_lines)
        elif self.state is not None and self.expansion_level == ExpansionLevel.Vector:
            return representation_vector(self.state)
        elif self.state is not None and self.expansion_level == ExpansionLevel.Matrix:
            return representation_matrix(self.state)
        return str(self.uid)  # pragme: no cover

    def combine(self) -> None:
        """
        Combines the fock and polarization into one vector or matrix and
        stores it under self.composite_vector or self.composite_matrix appropriately
        """
        from photon_weave.state.fock import Fock
        from photon_weave.state.polarization import Polarization

        for s in [self.fock, self.polarization]:
            if s.measured:
                raise ValueError(
                    "Parts of this envelope have already been destructively measured,"
                    " cannot combine"
                )

        assert isinstance(self.fock.expansion_level, ExpansionLevel)
        assert isinstance(self.polarization.expansion_level, ExpansionLevel)
        if self.fock.expansion_level == ExpansionLevel.Label:
            self.fock.expand()
        if self.polarization.expansion_level == ExpansionLevel.Label:
            self.polarization.expand()

        while self.fock.expansion_level < self.polarization.expansion_level:
            self.fock.expand()

        while self.polarization.expansion_level < self.fock.expansion_level:
            self.polarization.expand()

        if (
            self.fock.expansion_level == ExpansionLevel.Vector
            and self.polarization.expansion_level == ExpansionLevel.Vector
        ):
            assert isinstance(self.fock, Fock)
            assert hasattr(self.fock, "state") and isinstance(
                self.fock.state, jnp.ndarray
            )
            assert isinstance(self.polarization, Polarization)
            assert isinstance(self.polarization.state, jnp.ndarray)

            self.state = jitted_kron(self.fock.state, self.polarization.state)
            self.expansion_level = ExpansionLevel.Vector
            self.fock.extract(0)
            self.polarization.extract(1)

        if (
            self.fock.expansion_level == ExpansionLevel.Matrix
            and self.polarization.expansion_level == ExpansionLevel.Matrix
        ):
            assert isinstance(self.fock.state, jnp.ndarray)
            assert isinstance(self.polarization.state, jnp.ndarray)
            # fock_broadcast = self.fock.state[:,:,None,None]
            # polarization_broadcast = self.polarization.state[None,None,:,:]
            # self.state = fock_broadcast * polarization_broadcast
            self.state = jitted_kron(self.fock.state, self.polarization.state)
            self.fock.extract(0)
            self.polarization.extract(1)
            self.expansion_level = ExpansionLevel.Matrix

    @property
    def composite_envelope(self) -> Union["CompositeEnvelope", None]:
        """
        Property, which return appropriate composite envelope
        """
        from photon_weave.state.composite_envelope import CompositeEnvelope

        if self.composite_envelope_id is not None:
            ce = CompositeEnvelope._instances[self.composite_envelope_id][0]
            assert ce is not None, "Composite Envelope should exist"
            return ce
        return None

    def set_composite_envelope_id(self, uid: uuid.UUID) -> None:
        self.composite_envelope_id = uid

    def expand(self, *args: Any, **kwargs: Any) -> None:
        """
        Expands the state.
        If state is in the Fock and Polarization instances
        it expands those

        Parameters
        ----------
        *args:  Tuple
            Unused argument for compatibility, with routing functionality
        **kwargs: Dict
            Unused argument for compatibility, with routing functionality

        """
        if self.state is None:
            self.fock.expand()
            self.polarization.expand()
        else:
            assert isinstance(self.expansion_level, ExpansionLevel)
            self.state, self.expansion_level = state_expand(
                self.state, self.expansion_level, self.dimensions
            )

    def measure(
        self,
        *states: "BaseState",
        separate_measurement: bool = False,
        destructive: bool = True,
    ) -> Dict["BaseState", int]:
        """
        Measures the envelope. If the state is measured partially, then the state are
        moved to their respective spaces. If the measurement is destructive, then the
        state is destroyed post measurement.

        Parameter
        ---------
        *states: Optional[BaseState]
            Optional, when measuring spaces individualy
        separate_measurement:bool
            if True given states will be measured separately and the state which is not
            measured will be preserved (False by default)
        destructive: bool
            If False, the measurement will not destroy the state after the measurement.
            The state will still be affected by the measurement (True by default)

        Returns
        -------
        Dict[BaseState,int]
            Dictionary of outcomes, where the state is key and its outcome measurement
            is the value (int)
        """
        if self.measured:
            raise ValueError("Envelope has already been destroyed")

        # Check if given states are part of this envelope
        for s in states:
            assert s in [self.fock, self.polarization]
        if len(states) == 0:
            states = (self.fock, self.polarization)

        outcomes = {}
        if self.state is None:
            for s in [self.polarization, self.fock]:
                out = s.measure()
                for k, v in out.items():
                    outcomes[k] = v
        else:
            self.reorder(self.fock, self.polarization)
            if not separate_measurement and len(states) == 1:
                states = (self.fock, self.polarization)
            if len(states) == 2:
                separate_measurement = True

            assert all(isinstance(s.index, int) for s in states)

            match self.expansion_level:
                case ExpansionLevel.Vector:
                    outcomes, self.state = measure_vector(
                        [self.fock, self.polarization], list(states), self.state
                    )
                case ExpansionLevel.Matrix:
                    outcomes, self.state = measure_matrix(
                        [self.fock, self.polarization], list(states), self.state
                    )

            # Post Measurement process
            C = Config()
            for s in [self.fock, self.polarization]:
                if separate_measurement and s not in states:
                    s.state = self.state
                    s.index = None
                    s.expansion_level = self.expansion_level
                    # Try to contract the state twice
                    if C.contractions:
                        s.contract()
                        s.contract()
                else:
                    s.state = jnp.zeros((s.dimensions, 1))
                    s.state = s.state.at[outcomes[s], 0].set(1)
                    s.expansion_level = ExpansionLevel.Vector
                    s.index = None
                    if destructive:
                        s._set_measured()
                    else:
                        while (
                            s.expansion_level > ExpansionLevel.Label
                        ) and C.contractions:
                            s.contract()

        if destructive:
            self._set_measured()

        return outcomes

    def measure_POVM(
        self,
        operators: List[jnp.ndarray],
        *states: "BaseState",
        destructive: bool = True,
    ) -> Tuple[int, Dict["BaseState", int]]:
        """
        Positive Operation-Valued Measurement
        The measured state is destroyed (discarded) by default. If not, then
        it is affected by the measurement, but not discarded. If measuring
        two spaces this method automatically combines the two spaces if
        not already combined.

        Parameters
        ----------
        operators: List[jnp.ndarray]
            List of the POVM operators
        *states:Tuple[Union[np.ndarray, jnp.ndarray],
                     Optional[Union[np.ndarray, jnp.ndarray]]
            States on which the POVM measurement should be carried out,
            Order of the states must resemble order of the operator tensoring
        destructive: bool
            If True then after the measurement the density matrix is discarded

        Returns
        -------
        int
            The index of the measurement corresponding to the outcome
        """
        from photon_weave.state.fock import Fock
        from photon_weave.state.polarization import Polarization

        if self.measured:
            raise ValueError("This envelope was already measured")

        # Check the validity of the given states
        if len(states) == 2:
            if (isinstance(states[0], Fock) and isinstance(states[1], Fock)) or (
                isinstance(states[0], Polarization)
                and isinstance(states[1], Polarization)
            ):
                raise ValueError("Given states have to be unique")
        elif len(states) > 2:
            raise ValueError("Too many states given")
        for s in states:
            if s is not self.polarization and s is not self.fock:
                raise ValueError(
                    "Given states have to be members of the envelope, "
                    "use env.fock and env.polarization"
                )

        # Handle partial uncombined measurement
        if len(states) == 1 and self.state is None:
            if isinstance(states[0], Fock):
                outcome = self.fock.measure_POVM(operators, destructive=destructive)
            elif isinstance(states[0], Polarization):
                outcome = self.polarization.measure_POVM(
                    operators, destructive=destructive
                )
            return outcome

        # Handle the measurement if the state is in composite envelope product state
        if self.composite_envelope_id is not None:
            assert self.composite_envelope is not None
            return self.composite_envelope.measure_POVM(operators, *states)

        # Expand to matrix state if not alreay in it
        assert isinstance(self.expansion_level, ExpansionLevel)
        while self.expansion_level < ExpansionLevel.Matrix:
            self.expand()

        if len(states) == 2 and self.state is None:
            self.combine()

        assert isinstance(self.fock.index, int) and isinstance(
            self.polarization.index, int
        )
        assert isinstance(self.state, jnp.ndarray)
        self.reorder(self.fock, self.polarization)
        outcome, self.state = measure_POVM_matrix(
            [self.fock, self.polarization], list(states), operators, self.state
        )
        if len(states) == 2:
            if destructive:
                self._set_measured()
                self.fock._set_measured()
                self.polarization._set_measured()
            return (outcome, {})
        else:
            if destructive:
                other_state = (
                    self.fock if states[0] is self.polarization else self.polarization
                )
                other_state.state = trace_out_matrix(
                    [self.fock, self.polarization], [other_state], self.state
                )
                other_state.index = None
                other_state.expansion_level = ExpansionLevel.Matrix
                other_state.contract()
                states[0]._set_measured()
                self._set_measured()
        return (outcome, {})

    def apply_kraus(
        self,
        operators: List[jnp.ndarray],
        *states: BaseState,
        validity_check: bool = True,
    ) -> None:
        """
        Apply Kraus operator to the envelope.

        Parameters
        ----------
        operators: List[Union[np.ndarray, jnp.ndarray]]
            List of Kraus operators
        states:
            List of states in the same order as the tensoring of operators
        validity_check: bool
            Checks if given channel is valid, True by default
        """
        from photon_weave.state.fock import Fock
        from photon_weave.state.polarization import Polarization

        if len(states) == 0:
            raise ValueError("At least one state must be defined")

        s_uids = [s.uid for s in states]
        if (len(states) == 2) and (
            (self.fock.uid not in s_uids) or (self.polarization.uid not in s_uids)
        ):
            raise ValueError("Both given states must belong to the Envelope")
        elif len(states) > 2:
            raise ValueError("Too many states given")

        # If any of the states is in bigger product state apply the kraus there
        if self.state is None and any(isinstance(s.index, tuple) for s in states):
            assert self.composite_envelope_id is not None
            assert isinstance(self.composite_envelope, CompositeEnvelope)
            self.composite_envelope.apply_kraus(operators, *states)
            return

        # Handle partial application
        if len(states) == 1 and self.state is None:
            if isinstance(states[0], Fock):
                self.fock.apply_kraus(operators)
            elif isinstance(states[0], Polarization):
                self.polarization.apply_kraus(operators)
            return

        # Combine the states if Kraus operators are applied to both states
        if self.state is None:
            self.combine()
        assert isinstance(self.state, jnp.ndarray)

        if validity_check:
            if not kraus_identity_check(operators):
                raise ValueError(
                    "Kraus operators do not sum to the identity sum K^dagg K != I"
                )

        # Reorder
        self.reorder(*states)

        state_objs: List[Optional[BaseState]] = [None, None]
        state_objs[self.fock.index] = self.fock  # type: ignore
        state_objs[self.polarization.index] = self.polarization  # type: ignore

        # Kraus operators are only applied to the density matrices
        while self.expansion_level < ExpansionLevel.Matrix:
            self.expand()
        self.state = apply_kraus_matrix(
            state_objs, states, self.state, operators  # type: ignore
        )

        C = Config()
        if C.contractions:
            self.contract()

    def reorder(self, *states: "BaseState") -> None:
        """
        Changes the order of states in the product state

        Parameters
        ----------
        *states: BaseState
            new order of states
        """
        states_list = list(states)
        if len(states_list) == 2:
            if (
                isinstance(states_list[0], Fock) and isinstance(states_list[1], Fock)
            ) or (
                isinstance(states_list[0], Polarization)
                and isinstance(states_list[1], Polarization)
            ):
                raise ValueError("Given states have to be unique")
        elif len(states_list) > 2:
            raise ValueError("Too many states given")

        for s in states_list:
            if s not in [self.polarization, self.fock]:
                raise ValueError(
                    "Given states have to be members of the envelope, "
                    "use env.fock and env.polarization"
                )

        if self.state is None:
            return

        # Creating the new order in the product space
        if len(states_list) == 1:
            if states_list[0] is self.fock:
                states_list.append(self.polarization)
            elif states_list[0] is self.polarization:
                states_list.append(self.fock)

        assert isinstance(self.fock.index, int)
        assert isinstance(self.polarization.index, int)
        assert isinstance(self.fock.dimensions, int)
        current_order: List[Optional[Any]] = [None, None]
        current_order[self.fock.index] = self.fock
        current_order[self.polarization.index] = self.polarization

        if current_order[0] is states_list[0] and current_order[1] is states_list[1]:
            return

        current_shape = [0, 0]
        current_shape[self.fock.index] = self.fock.dimensions
        current_shape[self.polarization.index] = 2

        if self.expansion_level == ExpansionLevel.Vector:
            assert isinstance(self.state, jnp.ndarray)
            assert self.state.shape == (self.dimensions, 1)
            tmp_vector = self.state.reshape((current_shape[0], current_shape[1]))
            tmp_vector = jnp.transpose(tmp_vector, (1, 0))
            self.state = tmp_vector.reshape(-1, 1)
            self.fock.index, self.polarization.index = (
                self.polarization.index,
                self.fock.index,
            )
        elif self.expansion_level == ExpansionLevel.Matrix:
            assert isinstance(self.state, jnp.ndarray)
            assert self.state.shape == (self.dimensions, self.dimensions)
            tmp_matrix = self.state.reshape(
                current_shape[0], current_shape[1], current_shape[0], current_shape[1]
            )
            tmp_matrix = jnp.transpose(tmp_matrix, (1, 0, 3, 2))
            self.state = tmp_matrix.reshape(
                (current_shape[0] * current_shape[1] for i in range(2))
            )
            self.fock.index, self.polarization.index = (
                self.polarization.index,
                self.fock.index,
            )

    def contract(
        self, final: ExpansionLevel = ExpansionLevel.Vector, tol: float = 1e-6
    ) -> None:
        """
        Attempts to contract the representation to the level defined in `final`argument.

        Parameters
        ----------
        final: ExpansionLevel
            Expected expansion level after contraction
        tol: float
            Tolerance when comparing matrices
        """
        # Will not attempt to contract past vector
        # final = ExpansionLevel.Vector
        assert isinstance(self.state, jnp.ndarray)
        if self.expansion_level == ExpansionLevel.Vector:
            return
        assert self.state.shape == (self.dimensions, self.dimensions)
        if self.expansion_level == ExpansionLevel.Matrix:
            self.state, self.expansion_level, success = state_contract(  # type: ignore
                self.state, self.expansion_level
            )

    def _set_measured(self, remove_composite: bool = True) -> None:
        if self.composite_envelope is not None and remove_composite:
            self.composite_envelope.envelopes.remove(self)
            self.composite_envelope_id = None
        self.measured = True
        self.state = None

    def trace_out(
        self, *states: BaseState
    ) -> Union[jnp.ndarray, int, "PolarizationLabel"]:
        """
        Traces out the system, returning only the given states,
        if states are not in the same product space, then they
        are combined.

        Parameters
        ----------
        *states: BaseState
            The given states will be returned in the given
            order (tensoring order), with the rest traced out
        """
        from photon_weave.state.fock import Fock
        from photon_weave.state.polarization import Polarization, PolarizationLabel

        if self.composite_envelope is not None:
            assert isinstance(self.composite_envelope, CompositeEnvelope)
            return self.composite_envelope.trace_out(*states)

        if self.state is None and self.composite_envelope is None:
            if len(states) == 1:
                assert isinstance(states[0], (Polarization, Fock))
                assert isinstance(
                    states[0].state, (int, PolarizationLabel, jnp.ndarray)
                )
                return states[0].state
            elif len(states) == 2:
                self.combine()

        self.reorder(self.fock, self.polarization)

        assert isinstance(self.fock.index, int)
        assert isinstance(self.polarization.index, int)
        assert self.state is not None

        match self.expansion_level:
            case ExpansionLevel.Vector:
                return trace_out_vector(
                    [self.fock, self.polarization], states, self.state
                )
            case ExpansionLevel.Matrix:
                return trace_out_matrix(
                    [self.fock, self.polarization], states, self.state
                )
        return jnp.ndarray([[1]])  # pragma: no cover

    def overlap_integral(self, other: Envelope, delay: float, n: float = 1) -> float:
        r"""
        Given delay in [seconds] this method computes overlap of temporal
        profiles between this envelope and other envelope.

        Args:
        self (Envelope): Self
        other (Envelope): Other envelope to compute overlap with
        delay (float): Delay of the `other`after self
        Returns:
        float: overlap factor
        """
        f1 = self.temporal_profile.get_function(
            t_a=0, omega_a=(C0 / n) / self.wavelength
        )
        f2 = other.temporal_profile.get_function(
            t_a=delay, omega_a=(C0 / n) / other.wavelength
        )
        integrand = lambda x: np.conj(f1(x)) * f2(x)
        result, _ = quad(integrand, -np.inf, np.inf)

        return result

    def resize_fock(self, new_dimensions: int, state: Optional[Fock] = None) -> bool:
        """
        Adjusts the dimension of the fock in the envelope. The dimensions are adjusted
        also when the fock space is in the product state.

        Parameters
        ----------
        new_dimensions: int
            New dimensions to resize to
        state: Optional[Fock]
            Optional Parameter, not needed, because there can only be one
            Fock state in an Envelope, but needed because of the routing
            logic

        Returns
        -------
        bool
            True if resizing succeeded
        """

        if self.state is None:
            return self.fock.resize(new_dimensions)

        reshape_shape = [-1, -1]
        assert isinstance(self.fock.dimensions, int)
        assert isinstance(self.fock.index, int)
        assert isinstance(self.polarization.dimensions, int)
        assert isinstance(self.polarization.index, int)
        reshape_shape[self.fock.index] = self.fock.dimensions
        reshape_shape[self.polarization.index] = self.polarization.dimensions

        if self.expansion_level == ExpansionLevel.Vector:
            assert isinstance(self.state, jnp.ndarray)
            assert self.state.shape == (self.dimensions, 1)
            reshape_shape.append(1)
            ps = self.state.reshape(reshape_shape)
            if new_dimensions > self.fock.dimensions:
                padding = new_dimensions - self.fock.dimensions
                pad_config = [(0, 0) for _ in range(ps.ndim)]
                pad_config[self.fock.index] = (0, padding)
                ps = jnp.pad(ps, pad_config, mode="constant", constant_values=0)
                self.state = ps.reshape(-1, 1)
                self.fock.dimensions = new_dimensions
                return True
            if new_dimensions < self.fock.dimensions:
                to = self.trace_out(self.fock)
                assert isinstance(to, jnp.ndarray)
                num_quanta = num_quanta_vector(to)
                if num_quanta >= new_dimensions:
                    # Cannot shrink because amplitues exist beyond new_dimensions
                    return False
                slices = [slice(None)] * ps.ndim
                slices[self.fock.index] = slice(0, new_dimensions)
                ps = ps[tuple(slices)]
                self.state = ps.reshape(-1, 1)
                self.fock.dimensions = new_dimensions
                return True
        if self.expansion_level == ExpansionLevel.Matrix:
            assert isinstance(self.state, jnp.ndarray)
            assert self.state.shape == (self.dimensions, self.dimensions)
            ps = self.state.reshape([*reshape_shape, *reshape_shape]).transpose(
                [0, 2, 1, 3]
            )
            if new_dimensions > self.fock.dimensions:
                padding = new_dimensions - self.fock.dimensions
                pad_config = [(0, 0) for _ in range(ps.ndim)]
                pad_config[self.fock.index * 2] = (0, padding)
                pad_config[self.fock.index * 2 + 1] = (0, padding)

                ps = jnp.pad(ps, pad_config, mode="constant", constant_values=0)
                self.fock.dimensions = new_dimensions
                ps = ps.transpose([0, 2, 1, 3])
                self.state = ps.reshape((self.dimensions, self.dimensions))
                return True
            if new_dimensions <= self.fock.dimensions:
                to = self.trace_out(self.fock)
                assert isinstance(to, jnp.ndarray)
                num_quanta = num_quanta_matrix(to)
                if num_quanta >= new_dimensions:
                    return False
                slices = [slice(None)] * ps.ndim
                slices[self.fock.index * 2] = slice(0, new_dimensions)
                slices[self.fock.index * 2 + 1] = slice(0, new_dimensions)

                ps = ps[tuple(slices)]
                ps = ps.transpose([0, 2, 1, 3])
                self.fock.dimensions = new_dimensions
                self.state = ps.reshape((self.dimensions, self.dimensions))
                return True
        return False  # pragma: no cover

    def apply_operation(
        self, operation: "Operation", *states: Union["Fock", "Polarization"]
    ) -> None:
        """
        Applies given operation to the correct state

        Parameters
        ----------
        operation: Operation
            Operation to be applied to the state
        state: Union[Fock, Polarization]
            The state to which the operator should be applied to
        """

        # Check that correct operation is applied to the correct system
        if isinstance(operation._operation_type, FockOperationType):
            if not isinstance(states[0], Fock):
                raise ValueError(
                    "Cannot apply {type(operation._operation_type)} to "
                    f"{type(states[0])}"
                )
        if isinstance(operation._operation_type, PolarizationOperationType):
            if not isinstance(states[0], Polarization):
                raise ValueError(
                    "Cannot apply {type(operation._operation_type)} to "
                    f"{type(states[0])}"
                )

        # If the system is held in a different state container,
        # then apply it there
        if self.state is None:
            states[0].apply_operation(operation)
            return

        if not jnp.any(jnp.abs(self.state) > 0):
            raise ValueError("The state is invalid." "The state only consists of 0s.")

        # Compute the operator
        if isinstance(operation._operation_type, FockOperationType) and isinstance(
            states[0], Fock
        ):
            to = self.fock.trace_out()
            assert isinstance(to, jnp.ndarray)
            operation.compute_dimensions(states[0]._num_quanta, to)
            states[0].resize(operation.dimensions[0])
        elif isinstance(
            operation._operation_type, PolarizationOperationType
        ) and isinstance(states[0], Polarization):
            # Given arguments 0,0 don't have an effect
            operation.compute_dimensions([0], jnp.array([0]))

        assert isinstance(self.fock.index, int)
        assert isinstance(self.fock.dimensions, int)
        assert isinstance(self.polarization.index, int)
        assert isinstance(self.polarization.dimensions, int)
        current_order_dict: Dict[int, BaseState] = {}
        current_order_dict[self.fock.index] = self.fock
        current_order_dict[self.polarization.index] = self.polarization
        current_order = [current_order_dict[i] for i in sorted(current_order_dict)]
        # current_order: List[Optional[BaseState]] = [None, None]
        # current_order[self.fock.index] = self.fock
        # current_order[self.polarization.index] = self.polarization
        match self.expansion_level:
            case ExpansionLevel.Vector:
                self.state = apply_operation_vector(
                    current_order, list(states), self.state, operation.operator
                )
            case ExpansionLevel.Matrix:
                self.state = apply_operation_matrix(
                    current_order, list(states), self.state, operation.operator
                )

        if not jnp.any(jnp.abs(self.state) > 0):
            raise ValueError(
                "The state is entirely composed of zeros, "
                "is |0⟩ attempted to be annihilated?"
            )
        if operation.renormalize:
            self.state = self.state / jnp.linalg.norm(self.state)
        C = Config()
        if C.contractions:
            self.contract()



from __future__ import annotations

from typing import TYPE_CHECKING, List, Tuple, Union

import jax.numpy as jnp

import photon_weave.extra.einsum_constructor as ESC

if TYPE_CHECKING:
    from photon_weave.state.base_state import BaseState


def trace_out_vector(
    state_objs: Union[List[BaseState], Tuple[BaseState, ...]],
    target_states: Union[List[BaseState], Tuple[BaseState, ...]],
    product_state: jnp.ndarray,
) -> jnp.ndarray:
    """
    Traces out the system, returning only the space of given target states

    Parameters
    ----------
    state_objs: Union[List[BaseState],Tuple[BaseState, ...]]
        List of the state objects in the product space
    target_states: Union[List[BaseState],Tuple[BaseState, ...]]
        List of the target states, which should be included
        in the resulting subspace
    product_state: jnp.ndarray
        Product state in the vector form

    Returns
    -------
    jnp.ndarray
        Traced out subspace including only the states given as
        target_states. The tensoring of thes spaces in the
        traced out subspace reflects the order of the spaces
        in the target_states.

    Notes
    -----
        Tensoring order in the product space must reflect the order
        of the states given in the state_objs argument.
    """
    state_objs = list(state_objs)
    target_states = list(target_states)

    total_dimensions = jnp.prod(jnp.array([s.dimensions for s in state_objs]))
    shape_dims = [s.dimensions for s in state_objs]
    shape_dims.append(1)

    assert product_state.shape == (total_dimensions, 1)

    product_state = product_state.reshape(shape_dims)

    einsum = ESC.trace_out_vector(state_objs, target_states)

    trace_out_state = jnp.einsum(einsum, product_state)

    return trace_out_state.reshape((-1, 1))


def trace_out_matrix(
    state_objs: Union[List[BaseState], Tuple[BaseState, ...]],
    target_states: Union[List[BaseState], Tuple[BaseState, ...]],
    product_state: jnp.ndarray,
) -> jnp.ndarray:
    """
    Traces out the system, returning only the space of given target states

    Parameters
    ----------
    state_objs: Union[List[BaseState],Tuple[BaseState, ...]]
        List of the state objects in the product space
    target_states: Union[List[BaseState],Tuple[BaseState, ...]]
        List of the target states, which should be included
        in the resulting subspace
    product_state: jnp.ndarray
        Product state in the vector form

    Returns
    -------
    jnp.ndarray
        Traced out subspace including only the states given as
        target_states. The tensoring of thes spaces in the
        traced out subspace reflects the order of the spaces
        in the target_states.

    Notes
    -----
        Tensoring order in the product space must reflect the order
        of the states given in the state_objs argument.
    """

    state_objs = list(state_objs)
    target_states = list(target_states)

    total_dimensions = jnp.prod(jnp.array([s.dimensions for s in state_objs]))
    shape_dimensions = [s.dimensions for s in state_objs] * 2
    new_dimensions = [jnp.prod(jnp.array([s.dimensions for s in target_states]))] * 2

    assert product_state.shape == (total_dimensions, total_dimensions)

    product_state = product_state.reshape(shape_dimensions)

    einsum = ESC.trace_out_matrix(state_objs, target_states)

    traced_out_state = jnp.einsum(einsum, product_state)

    return traced_out_state.reshape(new_dimensions)



from typing import TYPE_CHECKING

import jax.numpy as jnp

if TYPE_CHECKING:
    from photon_weave.operation import Operation


class FockDimensions:
    """
    An estimator for fock dimensions, used when Displace, Squeeze or Expression

    It works in the following way:
    1.) It gets some initial esimtation, based on num_quanta and operation type
    2.) Computes the expected amplitudes
    3.) Applies an operator and checks the amplitudes
    4.) If more than threshold percentage of state is contained in the
        resulting state then that dimensions are returned
    5.) If not, dimensions are increased and process runs again until
        Enough state is contained in the post operation state
    """

    def __init__(
        self,
        state: jnp.ndarray,
        operation: "Operation",
        num_quanta: int,
        threshold: float,
    ) -> None:
        self.state = state
        self.dimensions = state.shape[0]
        self.operation = operation
        self.threshold = threshold
        self.num_quanta = num_quanta

    def compute_dimensions(self) -> int:
        """
        Computes the dimensions needed for the current state.
        After the dimensions were determined they are returned.

        Returns
        -------
        int
            Always returns a positive integer, denoting number of dimensions needed
        """
        self._initial_estimate()
        while True:
            n = self._compute_dimensions()
            if n > 0:
                return n
            self._increase_dimensions(5)

    def _initial_estimate(self) -> None:
        """
        Initial estimate for the dimensions, implemented for Displace and Squeeze.
        If other type is tried, then current state dimensionality is used for the
        initial guess.
        """
        from photon_weave.operation.fock_operation import FockOperationType

        if self.operation._operation_type is FockOperationType.Displace:
            cutoff = int(
                self.num_quanta + 3 * jnp.abs(self.operation.kwargs["alpha"]) ** 2
            )
            if cutoff > self.dimensions:
                self._increase_dimensions(amount=int(cutoff) - self.dimensions)
        if self.operation._operation_type is FockOperationType.Squeeze:
            r = jnp.abs(self.operation.kwargs["zeta"])
            en = (2 * self.num_quanta + 1) * jnp.sinh(r) ** 2 + self.num_quanta
            en_int = int(jnp.ceil(en))
            cutoff = int(self.num_quanta + 3 * en_int)
            if cutoff > self.dimensions:
                self._increase_dimensions(amount=cutoff - self.dimensions)

    def _compute_dimensions(self) -> int:
        """
        Computes the dimensions needed with current dimension guess

        Returns
        -------
        int
            Number of dimensions required, if -1 then not enough dimensions
            were guessed. Dimensions should be increased then and this
            method needs to be tried again
        """
        if self.state.shape == (self.dimensions, 1):
            self.operation._dimensions = [self.dimensions]
            operator = self.operation._operation_type.compute_operator(
                [self.dimensions], **self.operation.kwargs
            )
            resulting_state = jnp.dot(operator, self.state)
            cdf: float = 0
            if resulting_state[-1, 0] > (1 - self.threshold) * 1e-3:
                return -1
            for i in range(len(resulting_state)):
                tmp = jnp.abs(resulting_state[i][0]) ** 2
                cdf += float(tmp)
                if cdf >= self.threshold:
                    return i + 3
            return -1
        if self.state.shape == (self.dimensions, self.dimensions):
            self.operation._dimensions = [self.dimensions]
            operator = self.operation._operation_type.compute_operator(
                [self.dimensions], **self.operation.kwargs
            )
            resulting_state = operator @ self.state @ operator.T.conj()
            cdf = 0
            if jnp.abs(resulting_state[-1, -1]) > (1 - self.threshold) * 1e-3:
                return -1
            for i in range(self.dimensions):
                tmp = jnp.abs(resulting_state[i, i])
                cdf += float(tmp)

                if cdf >= self.threshold:
                    return i + 3
            return -1
        return -1

    def _increase_dimensions(self, amount: int = 1) -> None:
        """
        Increases the dimensions of the state for the requested amount

        amount: int
            Amount of increase of the dimensions
        """
        if self.state.shape == (self.dimensions, 1):
            pad = jnp.zeros((amount, 1), dtype=self.state.dtype)
            # Vertically stack the padding with the current state
            self.state = jnp.vstack([self.state, pad])
            # Update the dimensions by the amount added
            self.dimensions += amount
        if self.state.shape == (self.dimensions, self.dimensions):
            pad_rows = jnp.zeros((amount, self.dimensions), dtype=self.state.dtype)
            pad_cols = jnp.zeros(
                (self.dimensions + amount, amount), dtype=self.state.dtype
            )

            self.state = jnp.vstack([self.state, pad_rows])

            self.state = jnp.hstack([self.state, pad_cols])

            self.dimensions += amount



from typing import Any, List, Optional, Union

import jax.numpy as jnp

from photon_weave.operation.composite_operation import CompositeOperationType
from photon_weave.operation.custom_state_operation import CustomStateOperationType
from photon_weave.operation.fock_operation import FockOperationType
from photon_weave.operation.polarization_operation import PolarizationOperationType
from photon_weave.state.expansion_levels import ExpansionLevel


class Operation:
    """
    Represents a quantum operation applied to a state in various frameworks (Fock,
    Polarization, Composite, or Custom). The `Operation` class provides functionalities
    for applying quantum operations, managing operator matrices, and computing
    dimensions for these operations. It is designed to handle custom and predefined
    operations, updating internal parameters based on the specific operation type.

    Attributes
    ----------
    _operator : Optional[jnp.ndarray]
        The matrix representation of the operation. This is only used for custom
        operations.
    _operation_type : Union[FockOperationType, PolarizationOperationType,
    CompositeOperationType, CustomStateOperationType]
        The type of the operation, which determines how the operation behaves.
    _apply_count : int
        Specifies how many times this operation will be applied. Defaults to 1.
    _renormalize : bool
        Determines if renormalization is required after applying the operation.
    kwargs : dict
        Additional keyword arguments required for the operation, such as specific
        parameters related to the operation type.
    _expansion_level : ExpansionLevel
        Expansion level required for the operation (inferred from the operation type).
    _expression : str
        String-based expression defining the operation if applicable (e.g., custom
        operations).
    _dimensions : List[int]
        The estimated dimensions required to apply the operation, computed based on the
        number of quanta and state.

    Methods
    -------
    __init__(operation_type, expression=None, apply_count=1, **kwargs)
        Initializes the operation, updates internal parameters based on the operation
        type, and validates required parameters.

    __repr__()
        Provides a string representation of the operation, including matrix formatting
        for custom operators.

    dimensions
        Getter and setter for the dimensions required for the operation.

    compute_dimensions(num_quanta, state)
        Computes the required dimensions for applying the operation based on the
        provided quanta and state.

    required_expansion_level
        Returns the required expansion level for the operation.

    renormalize
        Returns whether renormalization is required after applying the operation.

    operator
        Getter and setter for the operator matrix. This is computed automatically
        unless the operation type is custom.

    Raises
    ------
    KeyError
        If required parameters for the operation type are not provided.
    ValueError
        If an attempt is made to manually set the operator for non-custom operation
        types.
    AssertionError
        If the provided operator is not a valid `jnp.ndarray` for custom operations.
    """

    __slots__ = (
        "_operator",
        "_operation_type",
        "_apply_count",
        "_renormalize",
        "kwargs",
        "_expansion_level",
        "_expression",
        "_dimensions",
    )

    def __init__(
        self,
        operation_type: Union[
            FockOperationType,
            PolarizationOperationType,
            CompositeOperationType,
            CustomStateOperationType,
        ],
        expression: Optional[str] = None,
        apply_count: int = 1,
        **kwargs: Any,
    ) -> None:
        """
        Initializes an `Operation` instance with the specified type, expression, and
        application count.

        Parameters
        ----------
        operation_type : Union[FockOperationType, PolarizationOperationType,
        CompositeOperationType, CustomStateOperationType] The type of operation
        (e.g., Fock, Polarization, Composite, or Custom).
        expression : Optional[str], optional
            A string expression for custom operations, by default None.
        apply_count : int, optional
            The number of times to apply the operation, by default 1.
        **kwargs : Any
            Additional keyword arguments required for the operation type.

        Raises
        ------
        KeyError
            If any required parameter for the operation type is missing.
        AssertionError
            If the operator for a custom operation is not a valid `jnp.ndarray`.
        """
        self._operation_type: Union[
            FockOperationType,
            PolarizationOperationType,
            CompositeOperationType,
            CustomStateOperationType,
        ] = operation_type
        self._operator: Optional[jnp.ndarray] = None
        self._apply_count: int = apply_count
        self._renormalize: bool
        self.kwargs = kwargs

        self._operation_type.update(**kwargs)

        if operation_type is FockOperationType.Custom:
            assert isinstance(kwargs["operator"], jnp.ndarray)
            self._operator = kwargs["operator"]
            self._dimensions: List[int] = [self._operator.shape[0]]

        for param in operation_type.required_params:
            if param not in kwargs:
                raise KeyError(
                    f"The '{param}' argument is required for {operation_type.name}"
                )

    def __repr__(self) -> str:
        """
        Returns a string representation of the operation, including the operator matrix
        for custom operations.
        """
        if self._operator is None:
            repr_string = (
                f"{self._operation_type.__class__.__name__}.{self._operation_type.name}"
            )
        else:
            repr_string = f"{self._operation_type.__class__.__name__}"
            repr_string += f".{self._operation_type.name}\n"
            formatted_matrix: Union[str, List[str]]
            formatted_matrix = ""

            if self._operator is None:
                return repr_string

            for row in self._operator:
                formatted_row = "⎢ "  # Start each row with the ⎢ symbol
                for num in row:
                    formatted_row += f"{num.real:+.2f} "  # Include a space
                    # after the real part
                    # Add either "+" or "-" for the imaginary part based on the sign
                    if num.imag >= 0:
                        formatted_row += "+ "
                    else:
                        formatted_row += "- "

                    # Format the imaginary part and add "j"
                    formatted_row += f"{abs(num.imag):.2f}j   "

                formatted_row = formatted_row.strip() + " ⎥\n"
                formatted_matrix += formatted_row

            # Add top and bottom brackets
            formatted_matrix = formatted_matrix.strip().split("\n")
            formatted_matrix[0] = "⎡" + formatted_matrix[0][1:-1] + "⎤"
            formatted_matrix[-1] = "⎣" + formatted_matrix[-1][1:-1] + "⎦"
            formatted_matrix = "\n".join(formatted_matrix)

            repr_string = repr_string + formatted_matrix

        return repr_string

    @property
    def dimensions(self) -> List[int]:
        """
        Gets or sets the estimated dimensions required for applying the operation.

        Returns
        -------
        List[int]
            The list of required dimensions.
        """
        assert isinstance(self._dimensions, list)
        return self._dimensions

    @dimensions.setter
    def dimensions(self, dimensions: List[int]) -> None:
        """
        Sets the dimensions required for applying the operation.

        Parameters
        ----------
        dimensions : List[int]
            The dimensions to be set.
        """
        self._dimensions = dimensions

    def compute_dimensions(
        self,
        num_quanta: Union[int, List[int]],
        state: jnp.ndarray,
    ) -> None:
        """
        Computes and updates the required dimensions for applying this operation.

        Parameters
        ----------
        num_quanta : Union[int, List[int]]
            The current maximum number state amplitude or a list of values.
        state : jnp.ndarray
            The traced-out state(s) for dimension estimation.
        """
        if self._operation_type is not FockOperationType.Custom:
            self._dimensions = self._operation_type.compute_dimensions(
                num_quanta, state, **self.kwargs
            )

    @property
    def required_expansion_level(self) -> ExpansionLevel:
        """
        Returns the required expansion level for the operation.

        Returns
        -------
        ExpansionLevel
            The expansion level required.
        """
        return self._operation_type.required_expansion_level

    @property
    def renormalize(self) -> bool:
        """
        Indicates whether renormalization is required after applying the operation.

        Returns
        -------
        bool
            True if renormalization is required, False otherwise.
        """
        return self._operation_type.renormalize

    @property
    def operator(self) -> jnp.ndarray:
        """
        Gets or computes the operator matrix for the operation.

        Returns
        -------
        jnp.ndarray
            The matrix representing the operation.

        Raises
        ------
        AssertionError
            If the computed operator is not a valid `jnp.ndarray`.
        """
        self._operator = self._operation_type.compute_operator(
            self.dimensions, **self.kwargs
        )
        assert isinstance(self._operator, jnp.ndarray)
        return self._operator

    @operator.setter
    def operator(self, operator: jnp.ndarray) -> None:
        """
        Sets the operator matrix for custom operations only.

        Parameters
        ----------
        operator : jnp.ndarray
            The matrix to be set as the operator.

        Raises
        ------
        ValueError
            If the operation type is not custom.
        AssertionError
            If the provided operator is not a valid `jnp.ndarray`.
        """
        assert isinstance(operator, jnp.ndarray)
        if self._operation_type is not FockOperationType.Custom:
            raise ValueError("Operator can only be configured for the Custom types")
        self._operator = operator



import random
import sys
from typing import Any

import jax
import jax.numpy as jnp


class Config:
    _instance = None

    def __new__(cls, *args: Any, **kwargs: Any) -> "Config":
        if cls._instance is None:
            cls._instance = super(Config, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self) -> None:
        if not hasattr(self, "_initialized"):
            self._initialized = True  # Prevents reinitialization
            self._random_seed = random.randint(0, sys.maxsize)
            self._key = jax.random.PRNGKey(self._random_seed)
            self._contractions = True

    def set_seed(self, seed: int) -> None:
        """
        For reproducability one can set a seed for random operations
        Parameters
        ----------
        seed: int
            Seed to be used by random processes
        """
        self._random_seed = seed
        self._key = jax.random.PRNGKey(seed)

    @property
    def random_seed(self) -> int:
        return self._random_seed

    @property
    def random_key(self) -> jnp.ndarray:
        """
        Splits the current key and returns a new one for random operations
        """
        key, self._key = jax.random.split(self._key)
        return key

    def set_contraction(self, cs: bool) -> None:
        self._contractions = cs

    @property
    def contractions(self) -> bool:
        return self._contractions



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Tuple, Union, Callable, Optional

import jax
import jax.numpy as jnp

import photon_weave.extra.einsum_constructor as ESC
from photon_weave.photon_weave import Config

if TYPE_CHECKING:
    from photon_weave.state.base_state import BaseState


def measure_vector(
    state_objs: Union[List[BaseState], Tuple[BaseState, ...]],
    target_states: Union[List[BaseState], Tuple[BaseState, ...]],
    product_state: jnp.ndarray,
    prob_callback: Optional[Callable[[BaseState, jnp.ndarray], None]] = None
) -> Tuple[Dict[BaseState, int], jnp.ndarray]:
    """
    Measures state vector and returns the outcome with the
    post measurement state.

    Parameters
    ----------
    state_objs:Union[List[BaseState],Tuple[BaseState, ...]]
        List of all state objects which are in the probuct state
    target_states: Union[List[BaseState],Tuple[BaseState, ...]]
        List of the states, which should be measured
    product_state: jnp.ndarray
        Product state containing the states to be measured
    prob_callback: Optional[Callable[[BaseState, jnp.ndarray],None]]
        Probability callback, used for testing the correctness of
        measurement outcome probabilities

    Returns
    -------
    Tuple[Dict[BaseState, int], jnp.ndarray]
        A Tuple containing:
        - A dictionary mapping each target state to its measurement
        - A jax array representing the post measuremen state of the rest of the
          product state

    Notes:
    ------
    calls the prob_callback() if provided
    """
    state_objs = list(state_objs)
    assert isinstance(product_state, jnp.ndarray), \
        f"Expected jnp.ndarray, received {type(product_state)}"
    expected_dims = jnp.prod(jnp.array([s.dimensions for s in state_objs]))
    assert product_state.shape == (expected_dims, 1), \
       f"Expected shape: ({expected_dims},1) , Received shape {product_state.shape}"

    # Reshape the array into the tensor
    shape = [s.dimensions for s in state_objs]
    shape.append(1)
    product_state = product_state.reshape(shape)

    C = Config()
    outcomes = {}
    for idx, state in enumerate(target_states):
        probabilities = []
        dims = state.dimensions
        for moutcome in range(dims):
            slicer = [slice(None)] * len(product_state.shape)
            slicer[state_objs.index(state)] = moutcome
            slice_psi = product_state[tuple(slicer)]
            probabilities.append(
                jnp.sum(jnp.abs(slice_psi)**2)
                )
        probabilities = jnp.array(probabilities)
        probabilities /= jnp.sum(probabilities)

        # For testing correctness of measurement probabilities
        if prob_callback:
            prob_callback(state, probabilities)

        # Based on the probabilities and key we choose outcome
        key = C.random_key
        outcome = jax.random.choice(
            key, a=jnp.arange(state.dimensions), p=probabilities
        )
        outcomes[state] = int(outcome)

        # We remove the measured system from the product state
        indices: List[Union[slice, int]] = [slice(None)] * len(product_state.shape)
        indices[state_objs.index(state)] = outcomes[state]
        product_state = product_state[tuple(indices)]

        state_objs.remove(state)

    product_state = product_state.reshape(-1, 1)

    return outcomes, product_state


def measure_matrix(
    state_objs: Union[List[BaseState], Tuple[BaseState, ...]],
    target_states: Union[List[BaseState], Tuple[BaseState, ...]],
    product_state: jnp.ndarray,
    prob_callback: Optional[Callable[[BaseState, jnp.ndarray], None]] = None
) -> Tuple[Dict[BaseState, int], jnp.ndarray]:
    """
    Measures Density Matrix and returns the outcome with the
    post measurement state of the product states, which were not measured.

    Parameters
    ----------
    state_objs: List[BaseState]
        List of all state objects which are in the probuct state
    target_states: List[BaseState]
        List of the states, which should be measured
    product_state: jnp.ndarray
        Product state containing the states to be measured
    prob_callback: Optional[Callable[[BaseState, jnp.ndarray],None]]
        Probability callback, used for testing the correctness of
        measurement outcome probabilities

    Returns
    -------
    Tuple[Dict[BaseState, int], jnp.ndarray]
        A Tuple containing:
        - A dictionary mapping each target state to its measurement
        - A jax array representing the post measuremen state of the rest of the
          product state

    Notes:
    ------
    calls the prob_callback() if provided
    """
    state_objs = list(state_objs)

    assert isinstance(product_state, jnp.ndarray), \
        f"Expected jnp.ndarray, received {type(product_state)}"
    expected_dims = jnp.prod(jnp.array([s.dimensions for s in state_objs]))
    assert product_state.shape == (expected_dims, expected_dims), \
       f"Expected shape: ({expected_dims},{expected_dims}) , Received shape {product_state.shape}"

    # Reshape the array into the tensor
    shape = [s.dimensions for s in state_objs] * 2
    product_state = product_state.reshape(shape)

    C = Config()
    outcomes = {}
    for idx, state in enumerate(target_states):
        probabilities = []
        dims = state.dimensions
        state_index = state_objs.index(state)
        for moutcome in range(dims):
            slicer = [slice(None)] * len(product_state.shape)
            slicer[state_index] = moutcome
            slicer[state_index + len(state_objs)] = moutcome
            block = product_state[tuple(slicer)]
            block = block.reshape((int(block.size//block.size**0.5), -1))
            probabilities.append(
                float(jnp.real(jnp.trace(block)))
            )

        #probabilities = jnp.abs(jnp.diag(projected_state))
        probabilities = jnp.array(probabilities)
        probabilities /= jnp.sum(probabilities)
        # For testing correctness of measurement probabilities
        if prob_callback:
            prob_callback(state, probabilities)

        # Based on the probabilities and key we choose outcome
        key = C.random_key
        outcome = jax.random.choice(
            key, a=jnp.arange(state.dimensions), p=probabilities
        )
        outcomes[state] = int(outcome)

        # Construct post measurement state
        state_index = state_objs.index(state)
        row_idx = state_index
        col_idx = state_index + len(state_objs)
        indices = [slice(None)] * len(product_state.shape)
        indices[row_idx] = outcomes[state]  # type: ignore
        indices[col_idx] = outcomes[state]  # type: ignore
        product_state = product_state[tuple(indices)]

        state_objs.remove(state)

    if len(state_objs) > 0:
        new_dims = jnp.prod(jnp.array([s.dimensions for s in state_objs]))
        product_state = product_state.reshape((new_dims, new_dims))

    return outcomes, product_state


def measure_POVM_matrix(
    state_objs: Union[List[BaseState], Tuple[BaseState, ...]],
    target_states: Union[List[BaseState], Tuple[BaseState, ...]],
    operators: Union[List[jnp.ndarray], Tuple[jnp.ndarray]],
    product_state: jnp.ndarray,
) -> Tuple[int, jnp.ndarray]:
    """
    Peform POVM measurement on the given product state.

    Parameters
    ----------
    state_objs: List[BaseState]
        List of BaseState objects included in the product stats
    target_states: List[BaseState]
        List of the stater in the product state, which will
        be measured with given operators
    operators: List[jnp.ndarray]
        list of the POVM operators, which desctibe the POVM measurement.
        The operators need to have correct dimensionality.
    product_state: jnp.ndarray
        The product state on which the measurement will be performed

    Returns
    -------
    Tuple[int, jnp.ndarray]
        Returns tuple with the index of the measurement outcome, where
        index refers to the index of the operator given in the operators
        list and the post measurement state

    Notes
    -----
    The states need to be tensored in the correct order. This is handled
    by the Photon Weave by default. In case of using this function
    specifically, make sure to have the states correctly ordered.
    """

    state_objs = list(state_objs)
    target_states = list(target_states)
    # Transform the operators to tensors
    op_shape = [s.dimensions for s in target_states] * 2
    operators = [op.reshape(op_shape) for op in operators]
    product_state = product_state.reshape([s.dimensions for s in state_objs] * 2)
    target_dims = jnp.prod(jnp.array([s.dimensions for s in target_states]))

    einsum_op = ESC.apply_operator_matrix(state_objs, target_states)
    einsum_to = ESC.trace_out_matrix(state_objs, target_states)
    prob_list: List[float] = []

    # Get the dimensions
    for operator in operators:
        projected_state = jnp.einsum(
            einsum_op, operator, product_state, jnp.conj(operator)
        )

        traced_out_projected_state = jnp.einsum(einsum_to, projected_state).reshape(
            (target_dims, target_dims)
        )

        prob_list.append(float(jnp.trace(traced_out_projected_state).real))

    probabilities = jnp.array(prob_list)
    probabilities /= jnp.sum(probabilities)

    C = Config()
    key = C.random_key
    outcome = int(jax.random.choice(key, a=jnp.arange(len(operators)), p=probabilities))
    product_state = jnp.einsum(
        einsum_op, operators[outcome], product_state, jnp.conj(operators[outcome])
    )

    product_state /= jnp.linalg.norm(product_state)

    state_dims = jnp.prod(jnp.array([s.dimensions for s in state_objs]))
    product_state = product_state.reshape((state_dims, state_dims))
    return outcome, product_state



from typing import Callable, Dict, List

import jax.numpy as jnp
from jax.scipy.linalg import expm


def interpreter(
    expr: tuple,
    context: Dict[str, Callable[[List[int]], jnp.ndarray]],
    dimensions: List[int],
) -> jnp.ndarray:
    r"""
        Recursively compute an operator from an expression, context and dimensions

    Parameters
    ----------
        expr: tuple
            Expression defined in a lisp style tuples
        context: Dict[str, Callable[[List[int]], jnp.ndarray]]
            Context is a dictionary of callables (lambda functions)
            Every lambda should accept list of integers (dimensions)
            and return a jnp.ndarray operator matrix
        dimensions: List[int]
            List of dimensions, ordered in the same way as the states
            operated on

    Returns
    -------
        jnp.ndarray
            Correctly computed operator

    Notes
    -----
        When constructing operator which acts on multiple spaces, be
        carefull that you correctly kron the operators. Wrong order
        will not result in correct operator, but no exceptions will
        be produced.

    Usage
    -----
        Tuples are evaluated from the inner most tuple to the outer most
        one. The first element in a tuple is a `command`, followed by
        `arguments˙. There can be multiple arguments for some commands.
        Interpreter understands the following commands:
            - add: adds all arguments together, accepts any number of arguments
            >>> ('add', 1, 2, 3) -> Sums the arguments and is evaluated to 6
            - sub: substracts the arguments, accepts 2 arguments
            >>> ('sub', 3, 2) -> 1
            - s_mult: Scalar multiplication, accepts any number of arguments
            >>> ('s_mult', 1,2,A) -> 2*A, where A is a matrix
            - m_mult: Matrix multiplication, accepts and number of arguments
            >>> ('m_mult', A,B,C) -> A@B@C
            - div: Divides the values, accpets two arguments
            >>> ('div', A, B) -> A/B
            - kron: Kronecker multiplication, accepts and number of arguments
            >>> ('kron', A,B,C) -> jnp.kron(A,jnp.kron(B,C))
            - expm: Exponentiate matrix term, accepts one argument
            >>> ('expm', A) -> e^A

            Expressions can be nested to produce complex expressions:
            >>> ('expm', ('s_mult', 1j, jnp.pi, 'n'))

        Arguments can be numbers, arrays or matrices. Arguments can also
        be string types. If an argument is a string type, like in example 'n',
        then its value will be computed from the context.
            >>> context = {
            >>>    'n': lambda dims: number_operator(dims[0])
            >>>}
            In this case
            'n' will be computed as:
            >>> context['n'](dims)
        The index 0 tells the interpreter the which state this operator should
        correspond to. The list `dims` is passed by the interpreter and it contains, the
        list of dimensions of the state we wish to operate on.

    """
    if isinstance(expr, tuple):
        op, *args = expr
        if op == "add":
            result = interpreter(args[0], context, dimensions)
            for arg in args[1:]:
                result = jnp.add(result, interpreter(arg, context, dimensions))
            return result
        if op == "sub":
            result = interpreter(args[0], context, dimensions)
            result = jnp.subtract(result, interpreter(args[1], context, dimensions))
            return result
        elif op == "s_mult":
            result = interpreter(args[0], context, dimensions)
            for arg in args[1:]:
                result *= interpreter(arg, context, dimensions)
            return result
        elif op == "m_mult":
            result = interpreter(args[0], context, dimensions)
            for arg in args[1:]:
                result @= interpreter(arg, context, dimensions)
            return result
        elif op == "kron":
            result = interpreter(args[0], context, dimensions)
            for arg in args[1:]:
                result = jnp.kron(result, interpreter(arg, context, dimensions))
            return result
        elif op == "expm":
            return expm(interpreter(args[0], context, dimensions))
        elif op == "div":
            return interpreter(args[0], context, dimensions) / interpreter(
                args[1], context, dimensions
            )
    elif isinstance(expr, str):
        # Grab a value from the context
        return context[expr](dimensions)
    else:
        # Grab literal value
        return expr
    raise ValueError("Something went wrong in the expression interpreter!", expr)



from typing import List, Union

import jax
import jax.numpy as jnp
import numpy as np
from jax import jit
from jax.scipy.linalg import expm

jax.config.update("jax_enable_x64", True)
jitted_exp = jit(expm)


@jit
def identity_operator() -> jax.Array:
    """
    identity_operator _summary_

    :return: _description_
    :rtype: jax.Array
    """
    return jnp.eye(N=2)


@jit
def hadamard_operator() -> jax.Array:
    """
    hadamard_operator _summary_

    :return: _description_
    :rtype: jax.Array
    """
    return jnp.array([[1, 1], [1, -1]]) * (1 / jnp.sqrt(2))


@jit
def x_operator() -> jax.Array:
    """
    x_operator _summary_

    :return: _description_
    :rtype: jax.Array
    """
    return jnp.array([[0, 1], [1, 0]])


@jit
def y_operator() -> jax.Array:
    """
    y_operator _summary_

    :return: _description_
    :rtype: jax.Array
    """
    return jnp.array([[0, -1j], [1j, 0]])


@jit
def z_operator() -> jax.Array:
    """
    z_operator _summary_

    :return: _description_
    :rtype: jax.Array
    """
    return jnp.array([[1, 0], [0, -1]])


@jit
def s_operator() -> jax.Array:
    """
    s_operator _summary_

    :return: _description_
    :rtype: jax.Array
    """
    return jnp.array([[1, 0], [0, 1j]])


@jit
def t_operator() -> jax.Array:
    """
    t_operator _summary_

    :return: _description_
    :rtype: jax.Array
    """
    return jnp.array([[1, 0], [0, jnp.exp(1j * np.pi / 4)]])


@jit
def controlled_not_operator() -> jax.Array:
    """
    controlled_not_operator _summary_

    :return: _description_
    :rtype: jax.Array
    """
    return jnp.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])


@jit
def controlled_z_operator() -> jax.Array:
    """
    controlled_z_operator _summary_

    :return: _description_
    :rtype: jax.Array
    """
    return jnp.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]])


@jit
def swap_operator() -> jax.Array:
    """
    swap_operator _summary_

    :return: _description_
    :rtype: jax.Array
    """
    return jnp.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])


@jit
def sx_operator() -> jax.Array:
    """
    sx_operator _summary_

    :return: _description_
    :rtype: jax.Array
    """
    return jnp.array([[1 + 1j, 1 - 1j], [1 - 1j, 1 + 1j]]) / 2


@jit
def controlled_swap_operator() -> jax.Array:
    """
    controlled_swap_operator _summary_

    :return: _description_
    :rtype: jax.Array
    """
    return jnp.array(
        [
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1],
        ]
    )


@jit
def rx_operator(theta: float) -> jax.Array:
    """
    rx_operator _summary_

    :param theta: _description_
    :type theta: float
    :return: _description_
    :rtype: jax.Array
    """
    term_1 = jnp.cos(theta / 2)
    term_2 = 1j * jnp.sin(-theta / 2)
    return jnp.array([[term_1, term_2], [term_2, term_1]])


@jit
def ry_operator(theta: float) -> jax.Array:
    """
    ry_operator _summary_

    :param theta: _description_
    :type theta: float
    :return: _description_
    :rtype: jax.Array
    """
    term_1 = jnp.cos(theta / 2)
    term_2 = jnp.sin(theta / 2)
    return jnp.array([[term_1, -term_2], [term_2, term_1]])


@jit
def rz_operator(theta: float) -> jax.Array:
    """
    rz_operator _summary_

    :param theta: _description_
    :type theta: float
    :return: _description_
    :rtype: jax.Array
    """
    term1 = jnp.exp(-1j * theta / 2)
    term2 = jnp.exp(1j * theta / 2)
    return jnp.array([[term1, 0], [0, term2]])


@jit
def u3_operator(phi: float, theta: float, omega: float) -> jax.Array:
    """
    u3_operator _summary_

    :param phi: _description_
    :type phi: float
    :param theta: _description_
    :type theta: float
    :param omega: _description_
    :type omega: float
    :return: _description_
    :rtype: jax.Array
    """
    cos_term = jnp.cos(theta / 2)
    sin_term = jnp.sin(theta / 2)
    return jnp.array(
        [
            [cos_term, -jnp.exp(1j * omega) * sin_term],
            [
                jnp.exp(1j * phi) * sin_term,
                jnp.exp(1j * (phi + omega)) * cos_term,
            ],
        ]
    )


def annihilation_operator(cutoff: int) -> jnp.ndarray:
    """
    annihilation_operator _summary_

    :param cutoff: _description_
    :type cutoff: int
    :return: _description_
    :rtype: jnp.ndarray
    """
    return jnp.diag(jnp.sqrt(jnp.arange(1, cutoff, dtype=np.complex128)), 1)


def creation_operator(cutoff: int) -> jnp.ndarray:
    """
    creation_operator _summary_

    :param cutoff: _description_
    :type cutoff: int
    :return: _description_
    :rtype: jnp.ndarray
    """
    return jnp.conjugate(annihilation_operator(cutoff=cutoff)).T


def number_operator(cutoff: int) -> jnp.ndarray:
    """
    number_operator _summary_

    :param cutoff: _description_
    :type cutoff: int
    :return: _description_
    :rtype: jnp.ndarray
    """
    return jnp.matmul(creation_operator(cutoff), annihilation_operator(cutoff))


def squeezing_operator(cutoff: int, zeta: complex) -> jnp.ndarray:
    """
    squeezing_operator _summary_

    :param cutoff: _description_
    :type cutoff: int
    :param zeta: _description_
    :type zeta: complex
    :return: _description_
    :rtype: jnp.ndarray
    """
    create = creation_operator(cutoff=cutoff)
    destroy = annihilation_operator(cutoff=cutoff)
    operator = 0.5 * (jnp.conj(zeta) * (destroy @ destroy) - zeta * (create @ create))

    return jitted_exp(operator)


def displacement_operator(cutoff: int, alpha: complex) -> jnp.ndarray:
    """
    displacement_operator _summary_

    :param cutoff: _description_
    :type cutoff: int
    :param alpha: _description_
    :type alpha: complex
    :return: _description_
    :rtype: jnp.ndarray
    """
    create = creation_operator(cutoff=cutoff)
    destroy = annihilation_operator(cutoff=cutoff)
    operator = alpha * create - jnp.conj(alpha) * destroy
    return jitted_exp(operator)


def phase_operator(cutoff: int, theta: float) -> jnp.ndarray:
    r"""
    Returns a phase shift operator, given the dimensions

    .. math::
      \hat{R}(\theta) = \sum_{n=0}^{\text{cutoff}-1} e^{1 n \theta }|n\rangle \langle n|

    This operator applies a phase shift to each Fock state |n⟩ proportional to the
    integer n. The phase shift is given by :math:`e^{i n \theta}`, where `theta` is the
    phase shift parameter.


    Parameters
    ----------
    cutoff: int
        Cutoff dimensions
    theta: float
        Phase shift for the operator

    Returns
    -------
    jnp.ndarray
        Constructed opreator

    Notes
    -----
    The phase shift operator is unitary and is used to rotate the phase of a quantum
    state in the Fock basis. The diagonal matrix elements are complex exponentials that
    apply a phase proportional to the Fock state number.
    """
    indices = jnp.arange(cutoff)
    phases = jnp.exp(1j * indices * theta)
    return jnp.diag(phases)


# to do: implement beamsplitter here
@jit
def compute_einsum(
    einsum_str: str, *operands: Union[jax.Array, np.ndarray]
) -> jax.Array:
    """
    Computes einsum using the provided einsum_str and matrices
    with the gpu if accessible (jax.numpy).
    Parameters
    ----------
    einsum_str: str
        Einstein Sum String
    operands: Union[jax.Array, np.ndarray]
        Operands for the einstein sum
    Returns
    -------
    jax.Array
        resulting matrix after eintein sum
    """
    return jnp.einsum(einsum_str, *operands)


@jit
def apply_kraus(
    density_matrix: Union[np.ndarray, jnp.ndarray],
    kraus_operators: List[Union[np.ndarray, jnp.ndarray]],
) -> jnp.ndarray:
    """
    Apply Kraus operators to the density matrix.
    Parameters
    ----------
    density_matrix: Union[np.ndarray, jnp.ndarray]
        Density matrix onto which the Kraus operators are applied
    kraus_operators: List[Union[np.ndarray, jnp.ndarray]]
        List of Kraus operators to apply to the density matrix
    Returns
    jnp.ndarray
        density matrix after applying Kraus operators
    """
    new_density_matrix = jnp.zeros_like(density_matrix)
    for K in kraus_operators:
        new_density_matrix += K @ density_matrix @ jnp.conjugate(K).T
    return new_density_matrix

    # return sum(K @ density_matrix @ jnp.conjugate((K)).T for K in kraus_operators)


def kraus_identity_check(operators: List[jnp.ndarray], tol: float = 1e-6) -> bool:
    """
    Check if Kraus operator sum is less or equal to identity,
    thus representing a valid Kraus Channel

    Parameters
    ----------
    kraus_operators: List[np.ndarray, jnp.Array]
        List of the operators
    tol: float
        Tolerance for the floating-point comparisons

    Returns
    -------
    bool
        True if the Kraus operators sum to identity within the tolerance
    """
    # TODO check if we need dim and identity matrix
    # dim = operators[0].shape[0]
    # identity_matrix = jnp.eye(dim)

    kraus_sum = sum(jnp.matmul(jnp.conjugate(K.T), K) for K in operators)
    identity = jnp.eye(kraus_sum.shape[0])  # type: ignore
    is_valid = jnp.all(jnp.real(jnp.linalg.eigvals(identity - kraus_sum)) >= 0)
    return bool(is_valid)


@jit
def normalize_vector(vector: Union[jnp.ndarray, np.ndarray]) -> jnp.ndarray:
    """
    Normalizes the given vector and returns it
    Parameters
    ----------
    vector: Union[jnp.ndarray, np.ndarray]
        Vector which should be normalized

    Returns
    -------
    jnp.ndarray
        Normalized vector state
    """
    trace = jnp.trace(vector)
    return jnp.array(vector / trace)


@jit
def normalize_matrix(vector: Union[jnp.ndarray, np.ndarray]) -> jnp.ndarray:
    """
    Normalizes the given matrix and returns it

    Parameters
    ----------
    vector: Union[jnp.ndarray, np.ndarray]
        Vector which should be normalized

    Returns
    -------
    jnp.ndarray
        Normalized density matrix
    """
    norm = jnp.linalg.norm(vector)
    return vector / norm


def num_quanta_vector(vector: Union[jnp.ndarray, np.ndarray]) -> int:
    """
    Returns highest possible measurement outcome
    Parameters
    ----------
    vector: Union[jnp.ndarray, np.ndarray]
        vector for which the max possible quanta has to be calculated

    Returns
    -------
    int
        Highest possible measure outcome
    """
    non_zero_indices = jnp.nonzero(vector)[0]
    return int(non_zero_indices[-1])


def num_quanta_matrix(matrix: jnp.ndarray) -> jnp.int64:
    """
    Returns highest possible measurement outcome
    Parameters
    ----------
    matrix: Union[jnp.ndarray, np.ndarray]
        matrix for which the max possible quanta has to be calculated
    """
    non_zero_rows = jnp.any(matrix != 0, axis=1)
    non_zero_cols = jnp.any(matrix != 0, axis=0)

    highest_non_zero_index_row = (
        jnp.where(non_zero_rows)[0][-1].item() if jnp.any(non_zero_rows) else None
    )
    highest_non_zero_index_col = (
        jnp.where(non_zero_cols)[0][-1].item() if jnp.any(non_zero_cols) else None
    )
    assert highest_non_zero_index_row is not None  # for mypy but needs to be checked
    assert highest_non_zero_index_col is not None  # for mypy but needs to be checked
    # Determine the overall highest index
    highest_non_zero_index_matrix = max(
        highest_non_zero_index_row, highest_non_zero_index_col
    )
    return highest_non_zero_index_matrix



from enum import Enum
from typing import Any, List, Tuple, Union

import jax.numpy as jnp

from photon_weave._math.ops import (
    hadamard_operator,
    identity_operator,
    rx_operator,
    ry_operator,
    rz_operator,
    s_operator,
    sx_operator,
    t_operator,
    u3_operator,
    x_operator,
    y_operator,
    z_operator,
)
from photon_weave.state.expansion_levels import ExpansionLevel


class PolarizationOperationType(Enum):
    r"""
    PolarizationOperationType

    Constructs an operator, which acts on a single Polarization Space

    Identity (I)
    ------------
    Constructs Identity (:math:`\hat{I}`) operator
    .. math::

        \hat{I} = \begin{bmatrix} 1 & 0 \\ 0 & 1 \end{bmatrix}

    Pauli X (X)
    -----------
    Constructs Pauli X  (:math:`\hat{\sigma_X}`) operator

    .. math::

        \hat{\sigma_X} = \begin{bmatrix} 0 & 1 \\ 1 & 0 \end{bmatrix}

    Pauli Y (Y)
    -----------
    Constructs Pauli Y  (:math:`\hat{\sigma_Y}`) operator

    .. math::

        \hat{\sigma_y} = \begin{bmatrix}
            0 & -i \\
            i & 0
            \end{bmatrix}

    Pauli Z (Z)
    -----------
    Constructs Pauli Z  (:math:`\hat {\sigma_Z}`) operator

    .. math::

        \hat{\sigma_Z} = \begin{bmatrix} 1 & 0 \\ 0 & -1 \end{bmatrix}

    Hadamard (H)
    ------------
    Constructs Hadamard (:math:`\hat{H}`) operator

    .. math::

        \hat{H} = \frac{1}{\sqrt{2}} \begin{bmatrix} 1 & 1 \\ 1 & -1 \end{bmatrix}
    Phase Opeator (S)
    -----------------
    Constructs Phase (:math:`\hat S`) operator
    .. math::

        \hat{S} = \begin{bmatrix} 1 & 0 \\ 0 & i \end{bmatrix}

    T Operator (T)
    -------------
    Constructs T (:math:`\hat T`) operator
    .. math::
        \hat{T} = \begin{bmatrix} 1 & 0 \\ 0 & e^{i \pi/4} \end{bmatrix}

    Sqrt(X) Operator (SX)
    ---------------------
    Constructs SX (:math:`\hat{SX}`) operator
    .. math::

        \hat{SX} = \frac{1}{2} \begin{bmatrix} 1+i & 1-i \\ 1-i & 1+i \end{bmatrix}

    RX Operator (RX)
    ----------------
    Constructs RX (:math:`\hat{RX}(\theta)`) operator
    It rotates around X axis for given :math:`\theta` angle
    Requires an argument "theta" (:math:`\theta`)
    .. math::

        \hat{RX} = \begin{bmatrix}
        \cos\left(\frac{\theta}{2}\right) & -i\sin\left(\frac{\theta}{2}\right) \\
        -i\sin\left(\frac{\theta}{2}\right) & \cos\left(\frac{\theta}{2}\right)
        \end{bmatrix}

    Usage Example
    >>> op = Operation(PolarizationOperationType.RX, theta=jnp.pi)

    RY Operator (RY)
    ----------------
    Constructs RY (:math:`\hat{RY}(\theta)`) operator
    It rotates around Y axis for given :math:`\theta` angle
    Requires an argument "theta" (:math:`\theta`)

    .. math::

        \hat{RY} = \begin{bmatrix}
        \cos\left(\frac{\theta}{2}\right) & -\sin\left(\frac{\theta}{2}\right) \\
        \sin\left(\frac{\theta}{2}\right) & \cos\left(\frac{\theta}{2}\right)
        \end{bmatrix}

    Usage Example
    >>> op = Operation(PolarizationOperationType.RY, theta=jnp.pi)

    RZ Operator (RZ)
    ----------------
    Constructs RZ (:math:`\hat{RZ}(\theta)`) operator
    It rotates around Z axis for given :math:`\theta` angle
    Requires an argument "theta" (:math:`\theta`)
    .. math::

        \hat{RZ} = \begin{bmatrix}
        e^{-i\frac{\theta}{2}} & 0 \\
        0 & e^{i\frac{\theta}{2}}
        \end{bmatrix}

    Usage Example
    >>> op = Operation(PolarizationOperationType.RZ, theta=jnp.pi)

    U3 operator
    -----------
    Rotation with 3 Euler angles (:math:`\hat{U3}(\phi,\theta,\omega))`)
    Requires three arguments "phi", "theta", "omega"

    Usage Example
    >>> op = Operation(
        PolarizationOperationType.U3, phi=jnp.pi, theta=jnp.pi/2, omega=jnp.pi/4)

    Custom
    ------
    Constructs a custom operator, the operator needs to be given manually

    Usage Example
    >>> operator = jnp.array([[1,0],[0,1]])
    >>> op = Operation(PolarizationOperationType.Custom, operator=operator)
    """

    I: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        True,
        [],
        ExpansionLevel.Vector,
        1,
    )
    X: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        True,
        [],
        ExpansionLevel.Vector,
        2,
    )
    Y: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        True,
        [],
        ExpansionLevel.Vector,
        3,
    )
    Z: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        True,
        [],
        ExpansionLevel.Vector,
        4,
    )
    H: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        True,
        [],
        ExpansionLevel.Vector,
        5,
    )
    S: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        True,
        [],
        ExpansionLevel.Vector,
        6,
    )
    T: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        True,
        [],
        ExpansionLevel.Vector,
        7,
    )
    SX: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        True,
        [],
        ExpansionLevel.Vector,
        8,
    )
    RX: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        True,
        ["theta"],
        ExpansionLevel.Vector,
        9,
    )
    RY: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        True,
        ["theta"],
        ExpansionLevel.Vector,
        10,
    )
    RZ: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        True,
        ["theta"],
        ExpansionLevel.Vector,
        11,
    )
    U3: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        True,
        ["phi", "theta", "omega"],
        ExpansionLevel.Vector,
        12,
    )
    Custom: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        True,
        ["operator"],
        ExpansionLevel.Vector,
        13,
    )

    def __init__(
        self,
        renormalize: bool,
        required_params: List[str],
        required_expansion_level: ExpansionLevel,
        op_id: int,
    ) -> None:
        self.renormalize = renormalize
        self.required_params = required_params
        self.required_expansion_level = required_expansion_level

    def update(self, **kwargs: Any) -> None:
        """
        Empty method, doesn't do anything in PolarizationOperationType
        """
        return

    def compute_operator(self, dimensions: List[int], **kwargs: Any) -> jnp.ndarray:
        """
        Computes an operator

        Parameters
        ----------
        dimensions: List[int]
            Accepts dimensions list, but it does not have any effect
        **kwargs: Any
            Accepts the kwargs, where the parameters are passed

        Returns
        -------
        jnp.ndarray
            Returns operator matrix
        """
        match self:
            case PolarizationOperationType.I:
                return identity_operator()
            case PolarizationOperationType.X:
                return x_operator()
            case PolarizationOperationType.Y:
                return y_operator()
            case PolarizationOperationType.Z:
                return z_operator()
            case PolarizationOperationType.H:
                return hadamard_operator()
            case PolarizationOperationType.S:
                return s_operator()
            case PolarizationOperationType.T:
                return t_operator()
            case PolarizationOperationType.SX:
                return sx_operator()
            case PolarizationOperationType.RX:
                return rx_operator(kwargs["theta"])
            case PolarizationOperationType.RY:
                return ry_operator(kwargs["theta"])
            case PolarizationOperationType.RZ:
                return rz_operator(kwargs["theta"])
            case PolarizationOperationType.U3:
                return u3_operator(kwargs["phi"], kwargs["theta"], kwargs["omega"])
            case PolarizationOperationType.Custom:
                return kwargs["operator"]
        raise ValueError("Operator not recognized")

    def compute_dimensions(
        self,
        num_quanta: Union[int, List[int]],
        state: Union[jnp.ndarray, List[jnp.ndarray]],
        threshold: float = 1,
        **kwargs: Any,
    ) -> List[int]:
        """
        Computes required dimensions for the operator. In this
        case the required dimension is always 2.

        Parameters
        ----------
        num_quanta: int
            Accepts num quatna, but it doesn't have an effect
        state: jnp.ndarray
            Accepts traced out state of the state on which the operator
            will operate, doens't have an effect
        threshold: float
            Threshold value, doesn't have any effect
        **kwargs: Any
            Accepts parameters, but they do not have any effect

        Returns
        -------
        int
            Returns number of dimensions needed (always 2)
        """
        return [2]



from __future__ import annotations

# TODO
# import inspect
import uuid
from dataclasses import dataclass, field
from functools import reduce
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union, cast

import jax
import jax.numpy as jnp

# from photon_weave.extra.einsum_constructor import EinsumStringConstructor as ESC
import photon_weave.extra.einsum_constructor as ESC
from photon_weave._math.ops import (
    kraus_identity_check,
    num_quanta_matrix,
    num_quanta_vector,
)
from photon_weave.operation import (
    CustomStateOperationType,
    FockOperationType,
    Operation,
    PolarizationOperationType,
)
from photon_weave.photon_weave import Config
from photon_weave.state.expansion_levels import ExpansionLevel

from .utils.measurements import measure_matrix, measure_POVM_matrix, measure_vector
from .utils.operations import (
    apply_kraus_matrix,
    apply_operation_matrix,
    apply_operation_vector,
)
from .utils.state_transform import state_contract, state_expand
from .utils.trace_out import trace_out_matrix, trace_out_vector

# For static type checks
if TYPE_CHECKING:
    from photon_weave.state.base_state import BaseState  # pragma: no cover
    from photon_weave.state.custom_state import CustomState  # pragma: no cover
    from photon_weave.state.envelope import Envelope  # pragma: no cover
    from photon_weave.state.fock import Fock  # pragma: no cover


@jax.jit
def kron_reduce(arrays: list[jnp.ndarray]) -> jnp.ndarray:
    """
    Compute the Kronecker product of a sequence of arrays using JAX.

    Parameters:
    -----------
    arrays : list or tuple of jax.numpy.ndarray
        A sequence of 2D arrays (matrices) to compute the Kronecker product.

    Returns:
    --------
    jax.numpy.ndarray
        The Kronecker product of the input arrays as a single matrix.
        If `arrays` is empty, returns a 1x1 identity matrix.

    Notes:
    ------
    - The function uses `reduce` to iteratively compute the Kronecker product
      of all matrices in `arrays`.
    - JAX's just-in-time (jit) compilation is applied for optimization.
    - `jnp.kron` is used to compute the Kronecker product between two matrices.

    Example:
    --------
    arrays = [jnp.array([[1, 2], [3, 4]]), jnp.array([[0, 5], [6, 7]])]
    result = kron_reduce(arrays)
    print(result)
    # Output:
    # [[ 0  5  0 10]
    #  [ 6  7 12 14]
    #  [ 0 15  0 20]
    #  [18 21 24 28]]
    """
    return reduce(lambda a, b: jnp.kron(a, b), arrays, jnp.array([[1]]))


@dataclass(slots=True)
class ProductState:
    """
    Stores Product state and references to its constituents
    """

    expansion_level: ExpansionLevel
    container: CompositeEnvelopeContainer
    uid: uuid.UUID = field(default_factory=uuid.uuid4)
    state: jnp.ndarray = field(default_factory=lambda: jnp.array([[1]]))
    state_objs: List[BaseState] = field(default_factory=list)

    def __hash__(self) -> int:
        return hash(self.uid)

    def expand(self) -> None:
        """
        Expands the state from vector to matrix
        """
        if self.expansion_level < ExpansionLevel.Matrix:
            dims = int(jnp.prod(jnp.array([s.dimensions for s in self.state_objs])))
            self.state, self.expansion_level = state_expand(
                self.state, self.expansion_level, dims
            )
            for state in self.state_objs:
                state.expansion_level = ExpansionLevel.Matrix

    def contract(self, tol: float = 1e-6) -> None:
        """
        Attempts to contract the representation from matrix to vector

        Parameters
        ----------
        tol: float
            tolerance when comparing trace
        """
        if self.expansion_level is ExpansionLevel.Matrix:
            self.state, self.expansion_level, success = cast(
                tuple[jnp.ndarray, ExpansionLevel, bool],
                state_contract(self.state, self.expansion_level),
            )
            if success:
                for state in self.state_objs:
                    state.expansion_level = ExpansionLevel.Vector

    @property
    def size(self) -> int:
        """
        Returns the size of the product state

        Returns
        -------
        int
            The size of the product state in bytes
        """
        return self.state.nbytes

    def reorder(self, *ordered_states: "BaseState") -> None:
        """
        Changes the order of tensoring, all ordered states need to be given

        Parameters
        ----------
        *ordered_states: 'BaseState'
            States ordered in the new order
        """

        assert all(
            so in ordered_states for so in self.state_objs
        ), "All state objects need to be given"
        if self.expansion_level == ExpansionLevel.Vector:
            # Get the state and reshape it
            shape = [so.dimensions for so in self.state_objs]
            shape.append(1)
            state = self.state.reshape(shape)

            # Generate the Einsum String
            einsum = ESC.reorder_vector(self.state_objs, list(ordered_states))

            # Perform the reordering
            state = jnp.einsum(einsum, state)

            # Reshape and store the state
            self.state = state.reshape(-1, 1)

            # Update the new order
            self.state_objs = list(ordered_states)
        elif self.expansion_level == ExpansionLevel.Matrix:
            # Get the state and reshape it
            shape = [os.dimensions for os in self.state_objs] * 2
            state = self.state.reshape(shape)

            # Get the einstein sum string
            einsum = ESC.reorder_matrix(self.state_objs, list(ordered_states))

            # Perform reordering
            state = jnp.einsum(einsum, state)

            # Reshape and reorder self.state_objs to reflect the new order
            new_dims = jnp.prod(jnp.array([s.dimensions for s in ordered_states]))
            self.state = state.reshape((new_dims, new_dims))
            self.state_objs = list(ordered_states)

        # Update indices in all of the states in this product space
        self.container.update_all_indices()

    def measure(
        self,
        *states: "BaseState",
        separate_measurement: bool = False,
        destructive: bool = True,
    ) -> Dict["BaseState", int]:
        """
        Measures this subspace. If the state is measured partially, then the state
        are moved to their respective spaces. If the measurement is destructive, then
        the state is destroyed post measurement.

        Parameters
        ----------
        states: Optional[BaseState]
            Optional, when measuring spaces individualy
        separate_measurement:bool
            if True given states will be measured separately and the state which is not
            measured will be preserved (False by default)
        destructive: bool
            If False, the measurement will not destroy the state after the measurement.
            The state will still be affected by the measurement (True by default)

        Returns
        -------
        Dict[BaseState,int]
            Dictionary of outcomes, where the state is key and its outcome measurement
        is the value (int)
        """
        from photon_weave.state.polarization import Polarization, PolarizationLabel

        assert all(
            so in self.state_objs for so in states
        ), "All state objects need to be in product state"

        match self.expansion_level:
            case ExpansionLevel.Vector:
                outcomes, self.state = measure_vector(
                    self.state_objs, states, self.state
                )
            case ExpansionLevel.Matrix:
                outcomes, self.state = measure_matrix(
                    self.state_objs, states, self.state
                )
        # Handle post measurement processes
        for state in states:
            if destructive:
                state._set_measured()
            else:
                if isinstance(state, Polarization):
                    if outcomes[state] == 0:
                        state.state = PolarizationLabel.H
                    else:
                        state.state = PolarizationLabel.V
                else:
                    state.state = outcomes[state]
                state.expansion_level = ExpansionLevel.Label
                state.index = None

        self.state_objs = [s for s in self.state_objs if s not in states]
        C = Config()
        if C.contractions and len(self.state_objs) > 0:
            self.contract()
        return outcomes

    def measure_POVM(
        self,
        operators: List[jnp.ndarray],
        *states: "BaseState",
        destructive: bool = True,
    ) -> Tuple[int, Dict["BaseState", int]]:
        """
        Perform a POVM measurement.

        Parameters
        ----------
        operators: List[jnp.ndarray]
            List of POVM operators
        *stapartial=partials: BaseState
            List of states, on which the POVM measurement should be executed
            The order of the states must reflect the order of tensoring of
            individual hilbert space in the individual operator
        destructive: bool
            If desctructive is set to True, then the states will be
            destroyed post measurement

        Returns
        -------
        Tuple[int, Dict[Union['Fock', 'Polarization'], int]]
            Tuple, where the first element is the outcome of the POVM measurement
            and the other is a dictionary of outcomes if the measurement
            was desctructive and some states are not captured in the
            POVM operator measurements.
        """

        from photon_weave.state.custom_state import CustomState

        # Expand to matrix form if not already in matrix form
        if self.expansion_level == ExpansionLevel.Vector:
            self.expand()

        outcome, self.state = measure_POVM_matrix(
            self.state_objs, states, operators, self.state
        )

        other_outcomes = {}
        if destructive:
            # Custom State cannot be destroyed
            for s in states:
                if isinstance(s, CustomState):
                    s.state = trace_out_matrix(self.state_objs, [s], self.state)
                    s.expansion_level = ExpansionLevel.Matrix
                    s.index = None

            remaining_states = [s for s in self.state_objs if s not in states]
            if isinstance(
                CompositeEnvelope._instances[self.container.composite_uid], list
            ):
                other_outcomes = CompositeEnvelope._instances[
                    self.container.composite_uid
                ][0].measure(*states)
                for s in states:
                    del other_outcomes[s]
            self.state_objs = remaining_states

        C = Config()
        if C.contractions and len(self.state_objs) > 0:
            self.contract()
        return (outcome, other_outcomes)

    def apply_kraus(
        self,
        operators: Union[List[jnp.ndarray], Tuple[jnp.ndarray, ...]],
        *states: "BaseState",
    ) -> None:
        """
        Applies the Kraus oeprators to the selected states, called by the apply_kraus
        method in CopositeEnvelope

        Parameters
        ----------
        operators: operators:List[Union[np.ndarray, jnp.ndarray]]
            List of operators to be applied, must be tensored using kron
        *states:BaseState
            List of states to apply the operators to, the tensoring order in operators
            must follow the order of the states in this list
        """
        while self.expansion_level < ExpansionLevel.Matrix:
            self.expand()
        self.state = apply_kraus_matrix(
            self.state_objs, states, self.state, operators
            )
        C = Config()
        if C.contractions:
            self.contract()

    def trace_out(self, *states: "BaseState") -> jnp.ndarray:
        """
        Traces out the rest of the states from the product state and returns
        the resultint matrix or vector. If given states are in sparate product
        spaces it merges the product spaces.

        Parameters
        ----------
        *states: BaseState

        Returns
        -------
        jnp.ndarray
            Traced out system including only the requested states in tesored
            in the order in which the states are given
        """
        match self.expansion_level:
            case ExpansionLevel.Vector:
                return trace_out_vector(self.state_objs, list(states), self.state)
            case ExpansionLevel.Matrix:
                return trace_out_matrix(self.state_objs, list(states), self.state)

        return jnp.ndarray([[1]])

    @property
    def is_empty(self) -> bool:
        """
        Returns True if the product state is empty (it contains no sates)

        Returns
        -------
        bool: True if empty, False if not
        """
        if len(self.state_objs) == 0:
            return True
        return False

    def resize_fock(self, new_dimensions: int, fock: "Fock") -> bool:
        """
        Resizes the space to the new dimensions.
        If the dimensions are more, than the current dimensions, then
        it gets padded. If the dimensions are less then the current
        dimensions, then it checks if it can shrink the space.

        Parameters
        ----------
        new_dimensions: bool
            New dimensions to be set
        fock: Fock
            The fock state, for which the dimensions
            must be changed

        Returns
        -------
        bool
            True if the resizing was succesfull
        """
        if self.expansion_level == ExpansionLevel.Vector:
            shape = [so.dimensions for so in self.state_objs]
            shape.append(1)
            ps = self.state.reshape(shape)
            if new_dimensions > fock.dimensions:
                padding = new_dimensions - fock.dimensions
                pad_config = [(0, 0) for _ in range(ps.ndim)]
                assert isinstance(fock.index, tuple)
                pad_config[fock.index[1]] = (0, padding)
                ps = jnp.pad(ps, pad_config, mode="constant", constant_values=0)
                fock.dimensions = new_dimensions
                dims = jnp.prod(jnp.array([s.dimensions for s in self.state_objs]))
                self.state = ps.reshape((dims, 1))
                return True
            if new_dimensions < fock.dimensions:
                to = fock.trace_out()
                assert isinstance(to, jnp.ndarray)
                assert isinstance(fock.index, tuple)
                num_quanta = num_quanta_vector(to)
                if num_quanta >= new_dimensions:
                    return False
                slices = [slice(None)] * ps.ndim
                slices[fock.index[1]] = slice(0, new_dimensions)
                ps = ps[tuple(slices)]
                fock.dimensions = new_dimensions
                dims = jnp.prod(jnp.array([s.dimensions for s in self.state_objs]))
                self.state = ps.reshape((dims, 1))
                return True

        if self.expansion_level == ExpansionLevel.Matrix:
            shape = [so.dimensions for so in self.state_objs]
            transpose_pattern = [
                item
                for i in range(len(self.state_objs))
                for item in [i, i + len(self.state_objs)]
            ]
            ps = self.state.reshape([*shape, *shape]).transpose(transpose_pattern)
            if new_dimensions > fock.dimensions:
                assert isinstance(fock.index, tuple)
                padding = new_dimensions - fock.dimensions
                pad_config = [(0, 0) for _ in range(ps.ndim)]
                pad_config[fock.index[1] * len(self.state_objs)] = (0, padding)
                pad_config[fock.index[1] * len(self.state_objs) + 1] = (0, padding)
                ps = jnp.pad(ps, pad_config, mode="constant", constant_values=0)
                fock.dimensions = new_dimensions
                dims = jnp.prod(jnp.array([s.dimensions for s in self.state_objs]))
                ps = ps.transpose(transpose_pattern)
                self.state = ps.reshape((dims, dims))
                return True
            if new_dimensions < fock.dimensions:
                to = fock.trace_out()
                assert isinstance(to, jnp.ndarray)
                assert isinstance(fock.index, tuple)
                num_quanta = num_quanta_matrix(to)
                if num_quanta >= new_dimensions:
                    return False
                slices = [slice(None)] * ps.ndim
                slices[fock.index[1] * len(self.state_objs)] = slice(0, new_dimensions)
                slices[fock.index[1] * len(self.state_objs) + 1] = slice(
                    0, new_dimensions
                )
                ps = ps[tuple(slices)]
                fock.dimensions = new_dimensions
                ps = ps.transpose(transpose_pattern)
                dims = jnp.prod(jnp.array([s.dimensions for s in self.state_objs]))
                self.state = jnp.array(ps.reshape((dims, dims)))
                return True
        return False

    def apply_operation(self, operation: Operation, *states: "BaseState") -> None:
        """
        Apply operation to the given states in this product state

        Parameters
        ----------
        operation: Operation
            The operation which will be applied
        *states: BaseState
            The states in the correct order to which the operation
            will be applied
        """
        from photon_weave.operation import CompositeOperationType
        from photon_weave.state.custom_state import CustomState
        from photon_weave.state.fock import Fock
        from photon_weave.state.polarization import Polarization

        if not jnp.any(jnp.abs(self.state) > 0):
            raise ValueError("The state is invalid" "State 0")

        if isinstance(operation._operation_type, FockOperationType):
            assert isinstance(states[0], Fock)
            assert len(states) == 1
            to = states[0].trace_out()
            assert isinstance(to, jnp.ndarray)
            operation.compute_dimensions(states[0]._num_quanta, to)
            # TODO
            _ = states[0].resize(operation.dimensions[0])

        elif isinstance(operation._operation_type, PolarizationOperationType):
            assert isinstance(states[0], Polarization)
            assert len(states) == 1
            # Parameters doesn't have any effect
            operation.compute_dimensions(0, jnp.array([0]))
        elif isinstance(operation._operation_type, CustomStateOperationType):
            assert isinstance(states[0], CustomState)
            assert len(states) == 1
            to = states[0].trace_out()
            assert isinstance(to, jnp.ndarray)
            operation.compute_dimensions(0, to)
        elif isinstance(operation._operation_type, CompositeOperationType):
            assert len(states) == len(
                operation._operation_type.expected_base_state_types
            )
            for i, s in enumerate(states):
                op_type = operation._operation_type
                assert isinstance(
                    s, op_type.expected_base_state_types[i]  # type: ignore
                )
            operation.compute_dimensions(
                [
                    s._num_quanta if isinstance(s, Fock) else s.dimensions
                    for s in states
                ],
                [s.trace_out() for s in states],  # type: ignore
            )
            for i, s in enumerate(states):
                if isinstance(s, Fock):
                    s.resize(operation._dimensions[i])

        match self.expansion_level:
            case ExpansionLevel.Vector:
                self.state = apply_operation_vector(
                    self.state_objs, states, self.state, operation.operator
                )
            case ExpansionLevel.Matrix:
                self.state = apply_operation_matrix(
                    self.state_objs, states, self.state, operation.operator
                )
        if not jnp.any(jnp.abs(self.state) > 0):
            raise ValueError(
                "The state is entirely composed of zeros, "
                "is |0⟩ attempted to be annihilated?"
            )
        if operation.renormalize:
            self.state = self.state / jnp.linalg.norm(self.state)

        # Reduce unneeded dimensionality of Fock spaces
        for so in self.state_objs:
            if isinstance(so, Fock):
                match so.expansion_level:
                    case ExpansionLevel.Vector:
                        num_quanta = max(1, num_quanta_vector(so.trace_out()))
                    case ExpansionLevel.Matrix:
                        num_quanta = max(1, num_quanta_matrix(so.trace_out()))
                so.resize(num_quanta + 1)
        C = Config()
        if C.contractions:
            self.contract()


@dataclass(slots=True)
class CompositeEnvelopeContainer:
    composite_uid: uuid.UUID
    envelopes: List["Envelope"] = field(default_factory=list)
    state_objs: List["BaseState"] = field(default_factory=list)
    states: List[ProductState] = field(default_factory=list)

    def append_states(self, other: "CompositeEnvelopeContainer") -> None:
        """
        Appends the states of two composite envelope containers
        Parameters
        ----------
        other: CompositeEnvelopeContainer
            Other composite envelope container
        """
        assert isinstance(other, CompositeEnvelopeContainer)
        self.states.extend(other.states)
        self.envelopes.extend(other.envelopes)

    def remove_empty_product_states(self) -> None:
        """
        Checks if a product state is empty and if so
        removes it
        """
        for state in self.states[:]:
            if state is not None:
                if state.is_empty:
                    self.states.remove(state)

    def update_all_indices(self) -> None:
        """
        Updates all of the indices of the state_objs
        """
        for state_index, state in enumerate(self.states):
            for i, so in enumerate(state.state_objs):
                if so is not None:
                    so.extract((state_index, i))
                    so.composite_envelope = CompositeEnvelope._instances[
                        self.composite_uid
                    ][0]


class CompositeEnvelope:
    """
    Composite Envelope is a pointer to a container, which includes the state
    Multiple Composite envelopes can point to the same containers.
    """

    _containers: Dict[Union["str", uuid.UUID], CompositeEnvelopeContainer] = {}
    _instances: Dict[Union["str", uuid.UUID], List["CompositeEnvelope"]] = {}

    def __init__(self, *states: Union["CompositeEnvelope", "Envelope", "CustomState"]):
        from photon_weave.state.base_state import BaseState
        from photon_weave.state.custom_state import CustomState
        from photon_weave.state.envelope import Envelope
        from photon_weave.state.fock import Fock
        from photon_weave.state.polarization import Polarization

        self.uid = uuid.uuid4()
        # Check if there are composite envelopes in the argument list
        composite_envelopes: List[CompositeEnvelope] = [
            e for e in states if isinstance(e, CompositeEnvelope)
        ]
        envelopes: List[Envelope] = [e for e in states if isinstance(e, Envelope)]
        state_objs: List[Union[CustomState, Fock, Polarization, BaseState]] = []
        for e in states:
            if isinstance(e, CustomState):
                state_objs.append(e)
            elif isinstance(e, Envelope):
                state_objs.append(e.fock)
                state_objs.append(e.polarization)
        for e in envelopes:
            if (
                e.composite_envelope is not None
                and e.composite_envelope not in composite_envelopes
            ):
                assert isinstance(
                    e.composite_envelope, CompositeEnvelope
                ), "e.composite_envelope should be CompositeEnvelope type"
                composite_envelopes.append(e.composite_envelope)

        ce_container = None
        for ce in composite_envelopes:
            assert isinstance(
                ce, CompositeEnvelope
            ), "ce should be CompositeEnvelope type"
            state_objs.extend(ce.state_objs)
            if ce_container is None:
                ce_container = CompositeEnvelope._containers[ce.uid]
            else:
                ce_container.append_states(CompositeEnvelope._containers[ce.uid])
            ce.uid = self.uid
        if ce_container is None:
            ce_container = CompositeEnvelopeContainer(self.uid)
        for e in envelopes:
            if e not in ce_container.envelopes:
                assert e is not None, "Envelope e should not be None"
                assert isinstance(e, Envelope), "e should be of type Envelope"
                ce_container.envelopes.append(e)
        for s in state_objs:
            if s.uid not in [x.uid for x in ce_container.state_objs]:
                ce_container.state_objs.append(s)

        CompositeEnvelope._containers[self.uid] = ce_container
        if not CompositeEnvelope._instances.get(self.uid):
            CompositeEnvelope._instances[self.uid] = []
        CompositeEnvelope._instances[self.uid].append(self)
        self.update_composite_envelope_pointers()

    def __repr__(self) -> str:
        return (
            f"CompositeEnvelope(uid={self.uid}, "
            f"envelopes={[e.uid for e in self.envelopes]}, "
            f"state_objects={[s.uid for s in self.state_objs]})"
        )

    @property
    def envelopes(self) -> List["Envelope"]:
        return CompositeEnvelope._containers[self.uid].envelopes

    @property
    def state_objs(self) -> List["BaseState"]:
        return CompositeEnvelope._containers[self.uid].state_objs

    @property
    def product_states(self) -> List[ProductState]:
        return CompositeEnvelope._containers[self.uid].states

    @property
    def container(self) -> CompositeEnvelopeContainer:
        return CompositeEnvelope._containers[self.uid]

    @property
    def states(self) -> List[ProductState]:
        return CompositeEnvelope._containers[self.uid].states

    def update_composite_envelope_pointers(self) -> None:
        """
        Updates all the envelopes to point to this composite envelope
        """
        for envelope in self.envelopes:
            envelope.set_composite_envelope_id(self.uid)

    def expand(self, *states: "BaseState") -> None:
        product_states = [
            p for p in self.states if any(so in p.state_objs for so in states)
        ]
        for p in product_states:
            p.expand()

    def contract(self, *state_objs: "BaseState") -> None:
        pass

    # @timing_decorator
    def combine(self, *state_objs: "BaseState") -> None:
        """
        Combines given states into a product state.

        Parameters
        ----------
        state_objs: BaseState
           Accepts many state_objs
        """
        from photon_weave.state.envelope import Envelope

        # stack = inspect.stack()
        # caller_frame = stack[1]
        # TODO
        # caller_function = caller_frame.function
        # caller_filename = caller_frame.filename
        # caller_line_number = caller_frame.lineno

        state_objs_set = set(state_objs)
        for ps in self.states:
            if state_objs_set.issubset(ps.state_objs):
                return  # Early exit on first match

        # Check if all states are included in composite envelope
        assert all(s in self.state_objs for s in state_objs)

        """
        Get all product states, which include any of the
        given states
        """
        existing_product_states = []
        for state in state_objs:
            for ps in self.product_states:
                if state in ps.state_objs:
                    existing_product_states.append(ps)
        # Removing duplicate product states
        # existing_product_states = list(set(existing_product_states))
        existing_product_states = list(dict.fromkeys(existing_product_states))

        """
        Ensure all states have the same expansion levels
        """
        minimum_expansion_level = ExpansionLevel.Vector
        for obj in state_objs:
            assert isinstance(obj.expansion_level, ExpansionLevel)
            if obj.expansion_level > minimum_expansion_level:
                minimum_expansion_level = ExpansionLevel.Matrix
                break

        # Expand the product spaces
        for product_state in existing_product_states:
            while product_state.expansion_level < minimum_expansion_level:
                product_state.expand()

        for obj in state_objs:
            if obj.index is None:
                assert isinstance(obj.expansion_level, ExpansionLevel)
                while obj.expansion_level < minimum_expansion_level:
                    obj.expand()
            elif isinstance(obj.index, int) and hasattr(obj, "envelope"):
                assert isinstance(obj.expansion_level, ExpansionLevel)
                while obj.expansion_level < minimum_expansion_level:
                    assert isinstance(obj.envelope, Envelope)
                    obj.expand()
        """
        Assemble all of the density matrices,
        and compile the indices in order
        """

        kron_arrays = []
        state_order = []

        # Collect arrays from new existing_product_states
        for product_state in existing_product_states:
            kron_arrays.append(product_state.state)
            state_order.extend(product_state.state_objs)
            product_state.state_objs = []
            product_state.state = jnp.array([[1]])

        # Collect arrays from new state objects
        for so in state_objs:
            if (
                hasattr(so, "envelope")
                and so.envelope is not None
                and so.index is not None
                and so.envelope.state is not None
                and not isinstance(so.index, tuple)
            ):
                kron_arrays.append(so.envelope.state)
                indices: List[Optional["BaseState"]] = [None, None]
                if isinstance(so.envelope.fock.index, int):
                    indices[so.envelope.fock.index] = so.envelope.fock
                if isinstance(so.envelope.polarization.index, int):
                    indices[so.envelope.polarization.index] = so.envelope.polarization
                state_order.extend([index for index in indices if index is not None])
                so.envelope.state = None
            elif so.index is None:
                assert isinstance(so.state, jnp.ndarray)
                kron_arrays.append(so.state)
                state_order.append(so)
                so.state = None

        state_vector_or_matrix = kron_reduce(kron_arrays)
        """
        Create a new product state object and append it to the states
        """
        ps = ProductState(
            expansion_level=minimum_expansion_level,
            container=self._containers[self.uid],
            state=state_vector_or_matrix,
            state_objs=state_order,
        )

        CompositeEnvelope._containers[self.uid].states.append(ps)

        """
        Remove empty states
        """
        self.container.remove_empty_product_states()
        self.container.update_all_indices()

    def reorder(self, *ordered_states: "BaseState") -> None:
        """
        Changes the order of the states in the produce space
        If not all states are given, the given states will be
        put in the given order at the beginnig of the product
        states

        Parameters
        ----------
        *ordered_states: BaseState
            ordered list of states
        """
        # Check if given states are shared in a product space
        states_are_combined = False
        for ps in self.states:
            if all(s in ps.state_objs for s in ordered_states):
                states_are_combined

        # TODO
        # stack = inspect.stack()

        # caller_frame = stack[1]

        # caller_function = caller_frame.function
        # caller_filename = caller_frame.filename
        # caller_line_number = caller_frame.lineno
        if not states_are_combined:
            self.combine(*ordered_states)

        # Get the correct product state:
        ps = [
            p for p in self.states if all(so in p.state_objs for so in ordered_states)
        ][0]

        # Create order
        new_order = [s for s in ps.state_objs]
        for i, ordered_state in enumerate(ordered_states):
            if new_order.index(ordered_state) != i:
                tmp = new_order[i]
                old_idx = new_order.index(ordered_state)
                new_order[i] = ordered_state
                new_order[old_idx] = tmp
        ps.reorder(*new_order)

    def measure(
        self,
        *states: "BaseState",
        separate_measurement: bool = True,
        destructive: bool = True,
    ) -> Dict["BaseState", int]:
        """
        Measures subspace in this composite envelope. If the state is measured
        partially, then the state are moved to their respective spaces. If the
        measurement is destructive, then the state is destroyed post measurement.

        Parameter
        ---------
        states: Optional[BaseState]
            States that should be measured
        separate_measurement:bool
            If true given states will be measured separately. Has only affect on the
            envelope states. If state is part of the envelope and separate_measurement
            is True, then the given state will be measured separately and the other
            state in the envelope won't be measured
        destructive: bool
            If False, the measurement will not destroy the state after the measurement.
            The state will still be affected by the measurement (True by default)

        Returns
        -------
        Dict[BaseState,int]
            Dictionary of outcomes, where the state is key and its outcome measurement
            is the value (int)
        """
        from photon_weave.state.envelope import Envelope
        from photon_weave.state.fock import Fock
        from photon_weave.state.polarization import Polarization

        outcomes: Dict["BaseState", int]
        outcomes = {}

        # Compile the complete list of states
        state_list = list(states)
        if not separate_measurement:
            for s in state_list:
                if hasattr(s, "envelope"):
                    assert isinstance(s.envelope, Envelope)
                    os: Union[Fock, Polarization]
                    if isinstance(s, Fock):
                        os = s.envelope.polarization
                    elif isinstance(s, Polarization):
                        os = s.envelope.fock
                    if os not in state_list:
                        state_list.append(os)

        # If the state resides in the BaseState or Envelope measure there
        for s in state_list:
            if isinstance(s.index, int) and hasattr(s, "envelope"):
                assert isinstance(s.envelope, Envelope)
                if separate_measurement:
                    out = s.envelope.measure(
                        s,
                        separate_measurement=separate_measurement,
                        destructive=destructive,
                    )
                else:
                    os = (
                        s.envelope.fock
                        if isinstance(s, Polarization)
                        else s.envelope.polarization
                    )
                    out = s.envelope.measure(
                        s,
                        os,
                        separate_measurement=separate_measurement,
                        destructive=destructive,
                    )
                for k, o in out.items():
                    outcomes[k] = o
            elif s.index is None:
                if not s.measured:
                    out = s.measure(
                        separate_measurement=separate_measurement,
                        destructive=destructive,
                    )
                    for k, o in out.items():
                        outcomes[k] = o

        # Measure in all of the product states
        product_states = [
            p for p in self.states if any(so in p.state_objs for so in state_list)
        ]

        for ps in product_states:
            ps_states = [so for so in state_list if so in ps.state_objs]
            out = ps.measure(
                *ps_states,
                separate_measurement=separate_measurement,
                destructive=destructive,
            )
            for key, item in out.items():
                outcomes[key] = item

        for s in state_list:
            if hasattr(s, "envelope") and destructive:
                assert isinstance(s.envelope, Envelope)
                s.envelope._set_measured()

        self._containers[self.uid].update_all_indices()
        self._containers[self.uid].remove_empty_product_states()
        return outcomes

    def measure_POVM(
        self,
        operators: List[jnp.ndarray],
        *states: "BaseState",
        destructive: bool = True,
    ) -> Tuple[int, Dict["BaseState", int]]:
        """
        Perform a POVM measurement.

        Parameters
        ----------
        operators: List[jnp.ndarray]
            List of POVM operators
        *states: List[Union[Fock, Polarization]]
            List of states, on which the POVM measurement should be executed
            The order of the states must reflect the order of tensoring of
            individual hilbert space in the individual operator
        destructive: bool
            If desctructive is set to True, then the states will be
            destroyed post measurement

        Returns
        -------
        int: Outcome result of index
        """

        # Check if the operator dimensions match
        dim = jnp.prod(jnp.array([s.dimensions for s in states]))
        for op in operators:
            if op.shape != (dim, dim):
                raise ValueError(
                    "At least on Kraus operator has incorrect dimensions: "
                    f"{op.shape}, expected({dim},{dim})"
                )

        # Get product states
        product_states = [
            p for p in self.states if any(so in p.state_objs for so in states)
        ]
        ps = None
        if len(product_states) > 1:
            all_states = [s for s in states]
            for p in product_states:
                all_states.extend([s for s in p.state_objs])
            self.combine(*all_states)
            ps = [p for p in self.states if any(so in p.state_objs for so in states)][0]
        elif len(product_states) == 0:
            self.combine(*states)
            ps = [p for p in self.states if any(so in p.state_objs for so in states)][0]
        else:
            ps = product_states[0]
        # Make sure the order of the states in tensoring is correct
        self.reorder(*states)

        outcome = ps.measure_POVM(operators, *states, destructive=destructive)
        return outcome

    def apply_kraus(
        self,
        operators: List[jnp.ndarray],
        *states: "BaseState",
        identity_check: bool = True,
    ) -> None:
        """
        Apply kraus operator to the given states
        The product state is automatically expanded to the density matrix
        representation. The order of tensoring in the operators should be the same
        as the order of the given states. The order of the states in the product
        state is changed to reflect the order of the given states.

        Parameters
        ----------
        operators: List[jnp.ndarray]
            List of all Kraus operators
        *states: BaseState
            List of the states, that the channel should be applied to
        identity_check: bool
            True by default, if true the method checks if kraus condition holds
        """

        from photon_weave.state.envelope import Envelope
        from photon_weave.state.fock import Fock
        from photon_weave.state.polarization import Polarization

        # Check the uniqueness of the states
        if len(states) != len(list(set(states))):
            raise ValueError("State list should contain unique elements")

        # Check if dimensions match
        dim = jnp.prod(jnp.array([s.dimensions for s in states]))
        for op in operators:
            if op.shape != (dim, dim):
                raise ValueError(
                    "At least on Kraus operator has incorrect dimensions: "
                    f"{op.shape}, expected({dim},{dim})"
                )

        # Check the identity sum
        if identity_check:
            if not kraus_identity_check(operators):
                raise ValueError("Kraus operators do not sum to the identity")
        # Get product states
        product_states = [
            p for p in self.states if any(so in p.state_objs for so in states)
        ]
        ps = None
        if len(product_states) > 1:
            all_states = [s for s in states]
            for p in product_states:
                all_states.extend([s for s in p.state_objs])
            self.combine(*all_states)
            ps = [p for p in self.states if any(so in p.state_objs for so in states)][0]
        elif len(product_states) == 0:
            # If only one state has to have kraus operators applied and the
            # state is not combined apply the kraus there
            if len(states) == 1:
                states[0].apply_kraus(operators)
                return
            elif len(states) == 2:
                if hasattr(states[0], "envelope") and hasattr(states[1], "envelope"):
                    if states[0].envelope == states[1].envelope:
                        assert isinstance(states[0].envelope, Envelope)
                        for s in states:
                            assert isinstance(s, (Fock, Polarization))
                        states[0].envelope.apply_kraus(operators, *states)
                        return
            self.combine(*states)
            ps = [p for p in self.states if any(so in p.state_objs for so in states)][0]
        else:
            ps = product_states[0]
        # Make sure the order of the states in tensoring is correct
        self.reorder(*states)

        ps.apply_kraus(operators, *states)

    def trace_out(self, *states: Union["BaseState"]) -> jnp.ndarray:
        """
        Traces out the rest of the states from the product state and returns
        the resultint matrix or vector. If given states are in sparate product
        spaces it merges the product spaces.

        Parameters
        ----------
        *states: Union[Fock, Polarization]

        Returns
        -------
        jnp.ndarray
            Traced out system including only the requested states in tesored
            in the order in which the states are given
        """
        product_states = [
            p for p in self.states if any(so in p.state_objs for so in states)
        ]
        assert len(product_states) > 0, "No product state found"
        ps: ProductState
        if len(product_states) > 1:
            all_states = [s for s in states]
            for p in product_states:
                all_states.extend([s for s in p.state_objs])
            self.combine(*all_states)
            product_states = [
                p for p in self.states if any(so in p.state_objs for so in states)
            ]
            assert (
                len(product_states) > 0
            ), "Only one product state should exist at this point"
        ps = product_states[0]

        if len(states) > 1:
            self.reorder(*states)

        return ps.trace_out(*states)

    def resize_fock(self, new_dimensions: int, fock: "Fock") -> bool:
        """
        Resizes the space to the new dimensions.
        If the dimensions are more, than the current dimensions, then
        it gets padded. If the dimensions are less then the current
        dimensions, then it checks if it can shrink the space.

        Parameters
        ----------
        new_dimensions: bool
            New dimensions to be set
        fock: Fock
            The fock state, for which the dimensions
            must be changed

        Returns
        -------
        bool
            True if the resizing was succesfull
        """
        from photon_weave.state.fock import Fock

        # Check if fock is Fock type
        if not isinstance(fock, Fock):
            raise ValueError("Only Fock spaces can be resized")

        # Check if fock is in this composite envelope
        if fock not in self.state_objs:
            raise ValueError(
                "Tried to resizing fock, which is not a part of this envelope"
            )

        if not isinstance(fock.index, tuple):
            return fock.resize(new_dimensions)

        ps = [ps for ps in self.product_states if fock in ps.state_objs]
        if len(ps) != 1:
            raise ValueError("Something went wrong")  # pragma : no cover

        new_ps = ps[0]
        self.reorder(fock)
        return new_ps.resize_fock(new_dimensions, fock)

    def apply_operation(
        self, operator: Operation, *states: Union["BaseState", "CustomState"]
    ) -> None:
        """
        Applies the operation to the correct product space. If operator
        has type CompositeOperator, then the product states are joined if
        the states are not yet in the same product state

        Parameters
        ----------
        operator: Operator
            Operator which should be appied to the state(s)
        states: BaseState
            States onto which the operator should be applied
        """
        if len(states) == 1:
            if not isinstance(states[0].index, tuple):
                assert hasattr(states[0], "apply_operation")
                states[0].apply_operation(operator)
                return
        product_states = [
            p for p in self.states if any(so in p.state_objs for so in states)
        ]
        ps = None
        if len(product_states) > 1 or (
            len(product_states) == 1
            and not all(state in product_states[0].state_objs for state in states)
        ):
            all_states = [s for s in states]
            for p in product_states:
                all_states.extend([s for s in p.state_objs])
            self.combine(*all_states)
            ps = [p for p in self.states if any(so in p.state_objs for so in states)][0]
        elif len(product_states) == 1:
            ps = product_states[0]
        if len(product_states) == 0:
            all_states = [s for s in states]
            self.combine(*all_states)
            ps = [p for p in self.states if any(so in p.state_objs for so in states)][0]

        assert isinstance(ps, ProductState)
        ps.apply_operation(operator, *states)
